#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import time
import uuid
import socket
import logging
from datetime                   import timedelta, datetime
from os.path                    import join, basename, expanduser
from utils                      import load_json, read_key, is_ip_private
from drm4g.utils.importlib      import import_module
from drm4g.managers.rocci       import CloudSetup
from drm4g.communicators        import REMOTE_JOBS_DIR
from drm4g                      import ( 
                                       COMMUNICATORS,
                                       REMOTE_VOS_DIR,
                                       DRM4G_DIR 
                                       )

logger = logging.getLogger(__name__)

cloud_setup_file = join(DRM4G_DIR, "etc", "cloudsetup.json")
#cloud_contextualisation_file = join(DRM4G_DIR, "etc", "cloud_config.conf")
generic_cloud_cfg = """
#cloud-config
users:
  - name: %s
    shell: /bin/bash
    sudo: ALL=(ALL) NOPASSWD:ALL
    lock-passwd: true+
    ssh-import-id: %s
    ssh-authorized-keys:
      - %s
"""

class Instance(object):

    def __init__(self, basic_data):
        self.data=basic_data
        self.id = None
        self.id_volume = None
        self.id_link = None
        self.int_ip = None
        self.ext_ip = None
        self.volume = int(basic_data['volume'])
        self.myproxy_server = basic_data.get('myproxy_server', '')
        self.private_key = expanduser(basic_data['private_key'])
        self.context_file = basename(self.private_key)+".login"
        self.vm_user = basic_data.get('vm_user', 'drm4g_admin')
        self.cloud_contextualisation_file = basic_data.get('vm_config', join(DRM4G_DIR, "etc", "cloud_config.conf"))
        self.comm = basic_data[ 'communicator' ]
        self.max_jobs_running = basic_data['max_jobs_running']
        self.vm_comm = basic_data.get('vm_communicator', self.comm)
        if self.vm_comm == 'local':
            self.vm_comm = 'pk_ssh'
        pub = read_key( self.private_key + ".pub" )

        try :
            cloud_setup = {}
            for name, features in load_json( cloud_setup_file ).items() :
                cloud_setup[ name ] =  CloudSetup(name, features)
        except Exception as err :
            logger.error( "Error reading the cloud setup file: " + str( err ) )

        infra_cfg = cloud_setup[ basic_data['lrms'] ]
        cloud_cfg = infra_cfg.cloud_providers[ basic_data['cloud_provider'] ]
        self.vo = infra_cfg.vo
        self.endpoint = cloud_cfg[ "endpoint" ]
        self.flavour = cloud_cfg[ "flavours" ][ basic_data['flavour'] ]
        self.app_name = basic_data['virtual_image']
        self.app = cloud_cfg[ "apps" ][ self.app_name ]

        communicator = import_module(COMMUNICATORS[ basic_data[ 'communicator' ] ] )
        com_obj = getattr( communicator , 'Communicator' ) ()
        com_obj.username       = basic_data['username']
        com_obj.frontend       = basic_data['frontend']
        com_obj.private_key    = self.private_key
        com_obj.public_key     = basic_data.get('public_key', self.private_key+'.pub')
        com_obj.work_directory = basic_data.get('scratch', REMOTE_JOBS_DIR)
        self.com_object = com_obj

        self.proxy_file = join( REMOTE_VOS_DIR , "x509up.%s" ) % self.vo

        '''
        commented so that the context file is created everytime
        just in case the user changed something in the contextualisation file
        #cmd = "ls %s" % self.context_file #to check if it exists
        out,err = self.com_object.execCommand( cmd )
        if err:
        '''
        with open( self.cloud_contextualisation_file, "r" ) as contex_file :
            cloud_config = contex_file.read()
            if 'vm_config' not in basic_data.keys():
                content = cloud_config % (self.vm_user, self.vm_user, pub)
            else:
                content = cloud_config
            logger.debug("Your contextualisation file %s :\n%s\n" % (self.cloud_contextualisation_file, content))
            #content = generic_cloud_cfg % (self.vm_user, self.vm_user, pub)
            cmd = "echo '%s' > %s" % (content, self.context_file)
            out,err = self.com_object.execCommand( cmd )
            if err:
                raise Exception("Wasn't able to create the context file %s." % self.context_file + err)

        cmd = "ls %s" % self.proxy_file #to check if it exists
        out,err = self.com_object.execCommand( cmd )
        if err:
            self._renew_voms_proxy()

    #This is here to avoid having the error "TypeError: can't pickle lock objects" when creating the pickled file
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['com_object']
        return odict

    def __setstate__(self, dict):
        self.__dict__.update(dict)
        communicator = import_module(COMMUNICATORS[ self.data[ 'communicator' ] ] )
        com_obj = getattr( communicator , 'Communicator' ) ()
        com_obj.username       = self.data['username']
        com_obj.frontend       = self.data['frontend']
        com_obj.private_key    = self.private_key
        com_obj.public_key     = self.data.get('public_key', self.private_key+'.pub')
        com_obj.work_directory = self.data.get('scratch', REMOTE_JOBS_DIR)
        self.com_object=com_obj

    def _exec_remote_cmd(self, command):
        logger.debug("~~~~~~~~~~~~~~~~ Going to execute remote command: ~~~~~~~~~~~~~~~~\n"+command)
        out, err = self.com_object.execCommand( command )
        logger.debug("~~~~~~~~~~~~~~~~         Command executed         ~~~~~~~~~~~~~~~~")
        return out, err

    def _renew_voms_proxy(self, cont=0):
        try:
            logger.debug( "Running rocci's _renew_voms_proxy function" )
            logger.debug( "_renew_voms_proxy count = %s" % str( cont ) )
            logger.error( "The proxy '%s' has probably expired" %  self.proxy_file )
            logger.info( "Renewing proxy certificate" )

            cmd = "rm %s" % self.proxy_file
            self.com_object.execCommand( cmd )
            if self.myproxy_server:
                LOCAL_X509_USER_PROXY = "X509_USER_PROXY=%s" % join ( REMOTE_VOS_DIR , self.myproxy_server )
            else :
                LOCAL_X509_USER_PROXY = "X509_USER_PROXY=%s/${MYPROXY_SERVER}" % ( REMOTE_VOS_DIR )
            cmd = "%s voms-proxy-init -ignorewarn " \
            "-timeout 30 -valid 24:00 -q -voms %s -noregen -out %s --rfc" % (
                LOCAL_X509_USER_PROXY ,
                self.vo ,
                self.proxy_file )

            out, err = self._exec_remote_cmd( cmd )
            self.log_output("_renew_voms_proxy", out, err)

            if err:
                logger.debug( "Ending  rocci's _renew_voms_proxy function with an error" )
                logger.error( "Error renewing the proxy(%s): %s" % ( cmd , err ) )
                raise Exception("Probably the proxy certificate hasn't been created. Be sure to run the the following command before trying again:" \
                    "\n    \033[93mdrm4g id <resource_name> init\033[0m")
            logger.info( "The proxy certificate will be operational for 24 hours" )
            logger.debug( "Ending  rocci's _renew_voms_proxy" )
        except socket.timeout:
            logger.debug("Captured a socket.time exception")
            if cont<3:
                self._renew_voms_proxy(cont+1)
            else:
                raise

    def create(self):
        logger.debug( "Running rocci's  create function" )
        if self.volume:
            self._create_volume()
            self._wait_storage()
        self._create_resource()
        logger.info( "Waiting until resource is active" )
        self._wait_compute()
        if self.volume :
            self._create_link()
        logger.debug( "Ending  rocci's create function" )

    def _wait_storage(self):
        logger.debug( "Running rocci's _wait_storage function" )
        now = datetime.now()
        end = now + timedelta( minutes = 60 )

        while now <= end :
            logger.debug( "  * _wait_storage - waited for %s minutes" % (timedelta( minutes = 60 ) - (end-now)) )
            out = self.get_description(self.id_volume)
            pattern = re.compile( "occi.storage.state\s*=\s*(.*)" )
            mo = pattern.findall( out )
            if mo and mo[ 0 ] == "online" :
                break
            time.sleep(10)
            now += timedelta( seconds = 10 )
        logger.debug( "Ending  rocci's _wait_storage function" )

    def _wait_compute(self):
        logger.debug( "Running rocci's _wait_compute function" )
        now = datetime.now()
        end = now + timedelta( minutes = 60 )

        while now <= end :
            logger.debug( "  * _wait_compute - waited for %s minutes" % (timedelta( minutes = 60 ) - (end-now)) )
            out = self.get_description(self.id)
            pattern = re.compile( "occi.compute.state\s*=\s*(.*)" )
            mo = pattern.findall( out )
            logger.debug( "The resource's state is %s" % mo[ 0 ] )
            if mo and mo[ 0 ] == "active" :
                break
            time.sleep(10)
            now += timedelta( seconds = 10 )
        logger.debug( "Ending  rocci's _wait_compute function" )

    def _create_link(self):
        logger.debug( "Running rocci's _create_link function" )
        logger.info( "Linking volume %s to resource %s" % (self.id_volume, self.id) )
        cmd = 'occi --endpoint %s --auth x509 --user-cred %s --voms --action link ' \
              '--resource %s -j %s' % (self.endpoint, self.proxy_file, self.id, self.id_volume )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("_create_link", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("_create_link 2", out, err)
        elif err :
            logger.error( "Ending  rocci's _create_link function with an error" )
            raise Exception( "Error linking resource and volume: %s" % out )
        self.id_link = out.rstrip('\n')
        logger.debug( "Ending  rocci's _create_link function" )

    def _create_resource(self):
        try:
            logger.debug( "Running rocci's _create_resource function" )
            logger.info( "Creating new resource" )
            cmd = 'occi --endpoint %s --auth x509 --user-cred %s --voms --action create --attribute occi.core.title="%s_DRM4G_VM_%s" ' \
                      '--resource compute --mixin %s --mixin %s --context user_data="file://$PWD/%s"' % (
                             self.endpoint, self.proxy_file, str(self.app_name).lower(), uuid.uuid4().hex, self.app, self.flavour, self.context_file )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("_create_resource", out, err)

            if 'certificate expired' in err :
                self._renew_voms_proxy()
                logger.debug( "After executing _renew_voms_proxy - Going to execute cmd again" )
                out, err = self._exec_remote_cmd( cmd )
                self.log_output("_create_resource 2", out, err)
            elif err :
                logger.error( "Ending  rocci's  _create_resource function with an error" )
                raise Exception( "Error creating VM : %s" % out )
            self.id = out.rstrip('\n')
            logger.info( "    Resource '%s' has been successfully created" % self.id )
            logger.debug( "Ending  rocci's  _create_resource function" )
        except Exception as err:
            raise Exception("Most likely the issue is being caused by a timeout error:\n"+str(err))

    def _create_volume(self):
        logger.debug( "Running rocci's _create_volume function" )
        logger.info( "Creating volume for resource %s" % self.id )
        cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action create --resource storage --attribute " \
              "occi.storage.size='num(%s)' --attribute occi.core.title=%s_DRM4G_Workspace_%s" % (
                     self.endpoint, self.proxy_file, str( self.volume ), str(self.app_name).lower(), uuid.uuid4().hex )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("_create_volume", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("_create_volume 2", out, err)
        elif err :
            logger.error( "Ending  rocci's _create_volume function with an error" )
            raise Exception( "Error creating volume : %s" % out )
        self.id_volume = out.rstrip('\n')
        logger.debug( "Ending  rocci's _create_volume function" )

    def delete(self):
        logger.debug( "Running rocci's  delete function" )
        logger.info( "Deleting resource %s" % self.id )
        if self.volume :
            cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action unlink --resource %s" % (
                             self.endpoint, self.proxy_file, self.id_link )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("delete (unlink)", out, err)

            if 'certificate expired' in err :
                self._renew_voms_proxy()
                logger.debug( "After executing _renew_voms_proxy - Going to execute cmd again" )
                out, err = self._exec_remote_cmd( cmd )
                self.log_output("delete (unlink) 2", out, err)
            elif err :
                logger.error( "Error unlinking volume '%s': %s" % ( self.id_volume, out ) )
            time.sleep( 20 )
            cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action delete --resource %s" % (
                             self.endpoint, self.proxy_file, self.id_volume )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("delete (volume)", out, err)

            if 'certificate expired' in err :
                self._renew_voms_proxy()
                logger.debug( "After executing _renew_voms_proxy - Going to execute cmd again" )
                out, err = self._exec_remote_cmd( cmd )
                self.log_output("delete (volume) 2", out, err)
            elif err :
                logger.error( "Error deleting volume '%s': %s" % ( self.id_volume, out ) )
        cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action delete --resource %s" % (
                             self.endpoint, self.proxy_file, self.id )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("delete (resource)", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            logger.debug( "After executing _renew_voms_proxy - Going to execute cmd again" )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("delete (resource) 2", out, err)
        elif err :
            logger.error( "Error deleting node '%s': %s" % ( self.id, out ) )
        logger.info( "    Resource '%s' has been successfully deleted" % self.id )
        logger.debug( "Ending  rocci's delete function" )

    def get_description(self, id):
        logger.debug( "Running rocci's  get_description function" )
        cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action describe --resource %s" % (
                             self.endpoint, self.proxy_file, id )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("get_description", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            logger.debug( "After executing _renew_voms_proxy - Going to execute cmd again" )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("get_description 2", out, err)
        elif err and "Insecure world writable dir" not in err:
            logger.error( "Ending  rocci's get_description function with an error" )
            raise Exception( "Error getting description node '%s': %s" % ( id, out ) )
        logger.debug( "Ending  rocci's get_description function" )
        return out

    def get_floating_ips(self):
        logger.debug( "Running rocci's  get_floating_ips function" )
        cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --dump-model | grep 'http://schemas.openstack.org/network/floatingippool'" % (
                      self.endpoint, self.proxy_file )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("get_floating_ips (floating check)", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("get_floating_ips (floating check) 2", out, err)
        elif err :
            logger.error( "Ending  rocci's get_floating_ips function with an error" )
        logger.debug( "Ending  rocci's get_floating_ips function" )
        return out

    def get_public_ip(self):
        logger.debug( "Running rocci's  get_public_ip function" )
        network_interfaces = self.get_network_interfaces()
        network_interface=""
        for n in network_interfaces[::-1] :
            if basename(n).lower() == 'public':
                network_interface = n
        if network_interface:
            cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action link --resource %s --link %s" % (
                      self.endpoint, self.proxy_file, self.id, network_interface )
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("get_public_ip", out, err)

            if 'certificate expired' in err :
                self._renew_voms_proxy()
                out, err = self._exec_remote_cmd( cmd )
                self.log_output("get_public_ip 2", out, err)

                if err:
                    raise Exception(str(err))
            elif err:
                raise Exception(str(err))
        else:
            floating_pools = self.get_floating_ips()
            if floating_pools:
                contents = floating_pools.split('\n')

                cont=0
                cond=False
                while cont<len(contents) and not cond:
                    items = contents[cont].split(';')
                    pairs = [item.split('=',1) for item in items]
                    d=dict((k,eval(v,{},{})) for (k,v) in pairs[1:])
                    mixin=d['scheme']+d['title']

                    cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action link --resource %s --link %s --mixin %s" % (
                          self.endpoint, self.proxy_file, self.id, network_interface, mixin )
                    out, err = self._exec_remote_cmd( cmd )
                    self.log_output("get_public_ip", out, err)

                    if 'No more floating ips in pool' in str(err):
                        cont+=1
                    elif 'certificate expired' in str(err) :
                        self._renew_voms_proxy()
                        continue
                    elif err:
                        logger.debug("An unexpected error occurred:\n"+str(err))
                        cont+=1
                    elif out:
                        #I'm assuming that out will have something that resembles this:
                        #http://stack-server-02.ct.infn.it:8787/network/interface/c6886e72-86bd-4a08-ab2b-f0769854a38a_90.147.16.53
                        #which is what the link commmand returns without that last "mixin" option
                        cond=True
                        logger.debug("\n\n\nI don't know if I will ever get to this point since I'm still not sure what's returned by the last executed command (since it's never worked until now)\n\n\n")
                    else:
                        logger.debug("There wasn't either an output or an error for the execution of the 'get_public_ip' function")
            else:
                raise Exception("Error trying to get a public IP")

        time.sleep( 10 )
        out = self.get_description(self.id)
        pattern = re.compile( "occi.networkinterface.address\s*=\s*(.*)" )
        mo = pattern.findall( out )
        if mo :
            for ip in mo :
                if is_ip_private( ip  ) :
                    self.int_ip = ip
                else :
                    self.ext_ip = ip
        if not self.ext_ip :
            logger.error( "Ending  rocci's get_public_ip function with an error" )
            raise Exception( "Error trying to get a public IP" )
        logger.debug( "Ending  rocci's get_public_ip function" )

    def get_network_interfaces(self):
        logger.debug( "Running rocci's  get_network_interfaces function" )
        cmd = "occi --endpoint %s --auth x509 --user-cred %s --voms --action list --resource network" % (
                             self.endpoint, self.proxy_file )
        out, err = self._exec_remote_cmd( cmd )
        self.log_output("get_network_interfaces", out, err)

        if 'certificate expired' in err :
            self._renew_voms_proxy()
            out, err = self._exec_remote_cmd( cmd )
            self.log_output("get_network_interfaces 2", out, err)
        elif err :
            logger.error( "Ending  rocci's get_network_interfaces function with an error" )
            raise Exception( "Error getting network list" )
        logger.debug( "Ending  rocci's get_network_interfaces function" )
        return out.strip().split()

    def get_ip(self):
        logger.debug( "Running rocci's  get_ip function" )
        logger.info( "Getting resource's IP direction" )
        out = self.get_description(self.id)
        pattern = re.compile( "occi.networkinterface.address\s*=\s*(.*)" )
        mo = pattern.findall( out )
        if mo :
            ip = mo[ 0 ]
            if not is_ip_private( ip ) :
                self.ext_ip = ip
                self.int_ip = ip
            else :
                self.get_public_ip()
            logger.info( "    Public IP: %s" % self.ext_ip )
        else :
            logger.error( "Ending  rocci's get_ip function with an error" )
            raise Exception( "Error getting IP" )

        logger.debug( "*********** get_ip -- self.int_ip ***********" )
        logger.debug( str(self.int_ip) )
        logger.debug( "*********** get_ip -- self.ext_ip ***********" )
        logger.debug( str(self.ext_ip) )
        logger.debug( "*********** get_ip -- ip ***********" )
        logger.debug( str(ip) )
        logger.debug( "*********** get_ip -- end ***********" )
        logger.debug( "Ending  rocci's get_ip function" )

    def log_output(self, msg, out, err, extra=None):
        logger.debug( "Command return:" )
        logger.debug( "*********** %s -- out ***********" % msg )
        logger.debug( str(out) )
        logger.debug( "*********** %s -- err ***********" % msg )
        logger.debug( str(err) )
        logger.debug( "*********** %s -- end ***********" % msg )



# ------------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------


import sys
import os
import shutil
import re
import logging
import arclib

logger=None
log_handler=None

re_src=re.compile(r"^gsiftp://.*?/~/\.gw_.*?_.*?/(.*)$")
re_dst=re.compile(r"^file://(/.*)$")

def run():
	global logger, log_handler
	logger=logging.getLogger("tm")
	log_handler=logging.FileHandler("/tmp/nordugrid.log")
	log_formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
	log_handler.setFormatter(log_formatter)
	logger.addHandler(log_handler)
	logger.setLevel(logging.INFO)
	
	process_line()
	
def stop():
	if logger:
		logger.removeHandler(log_handler)
		log_handler.close()

def process_line():
	line=sys.stdin.readline()
	
	tmp_line=line.split()
	action=""
	xfr_ids="-"
	cp_xfr_ids="-"
	mode="-"
	src_url="-"
	dst_url="-"
	
	if len(tmp_line)>0:
		action=tmp_line[0]
		
	if len(tmp_line)>1:
		xfr_ids=tmp_line[1]
		
	if len(tmp_line)>2:
		cp_xfr_ids=tmp_line[2]
		
	if len(tmp_line)>3:
		mode=tmp_line[3]

	if len(tmp_line)>4:
		src_url=tmp_line[4]
		
	if len(tmp_line)>5:
		dst_url=tmp_line[5]
	
	process_action(action, xfr_ids, cp_xfr_ids, mode, src_url, dst_url)
	
def action_init(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "INIT - - SUCCESS -"

def action_start(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "START %s - SUCCESS -" % xfr_ids
	
def action_end(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	# DELETE TEMPORAL DOWNLOAD DIRECTORY
	print "END %s - SUCCESS -" % xfr_ids

def action_rmdir(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "RMDIR %s - SUCCESS -" % xfr_ids

def action_exists(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "EXISTS %s - SUCCESS -" % xfr_ids

def action_mkdir(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "MKDIR %s - SUCCESS %s" % (xfr_ids, src_url)

def action_cp(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	res_src=re_src.match(src_url)
	res_dst=re_dst.match(dst_url)
	
	if res_src and res_dst:
		s=res_src.group(1)
		dst=res_dst.group(1)
		GW_LOCATION=os.environ['GW_LOCATION']
		src="%s/var/%s/nordugrid/%s" % (GW_LOCATION, xfr_ids, s)
		try:
			shutil.copy(src, dst)
			print "CP %s %s SUCCESS (%s->%s)" % (xfr_ids, cp_xfr_ids, src, dst)
		except Exception, e:
			print "CP %s %s FAILURE (%s->%s)" % (xfr_ids, cp_xfr_ids, src, dst)
	else:
		print "CP %s %s SUCCESS (%s->%s)" % (xfr_ids, cp_xfr_ids, src_url, dst_url)

def action_finalize(xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	print "FINALIZE - - SUCCESS -"
	sys.exit(0);

action_functions = {
	'INIT': 	action_init,
	'START': 	action_start,
	'END':		action_end,
	'RMDIR':	action_rmdir,
	'EXISTS':	action_exists,
	'MKDIR':	action_mkdir,
	'CP':		action_cp,
	'FINALIZE':	action_finalize
}

def process_action(action, xfr_ids, cp_xfr_ids, mode, src_url, dst_url):
	logger.info("GOT MESSAGE: \"%s\"" % " ".join([action, xfr_ids, cp_xfr_ids, mode, src_url, dst_url]))
	logger.info("  Action: %s" % action)
	logger.info("  xfr_ids: %s" % xfr_ids)
	logger.info("  Cp_xfr_ids: %s" % cp_xfr_ids)
	logger.info("  Mode: %s" % mode)
	logger.info("  Src_url: %s" % src_url)
	logger.info("  Dst_url: %s" % dst_url)
	
	tmp_action=action.upper()
	if action_functions.has_key(tmp_action):
		action_functions[tmp_action](xfr_ids, cp_xfr_ids, mode, src_url, dst_url)
		
	sys.stdout.flush()



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Configure DRM4G's daemon, scheduler and console_logger parameters.

Usage:
   drm4g conf ( daemon | sched | console_logger ) [ options ]

Options:
   -d --debug    Debug mode
"""

import os
from drm4g  import DRM4G_GWD_CONF, DRM4G_console_logger_CONF, DRM4G_SCHED_CONF, console_logger

def run( arg ) :
    if arg[ 'daemon' ] :
        conf_file = DRM4G_GWD_CONF
    elif arg[ 'console_logger' ]:
        conf_file = DRM4G_console_logger_CONF
    else :
        conf_file = DRM4G_SCHED_CONF
    console_logger.debug( "Editing '%s' file" % conf_file )
    os.system( "%s %s" % ( os.environ.get('EDITOR', 'nano') , conf_file ) )



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import pickle
import logging
from drm4g.utils.importlib import import_module
from drm4g.communicators   import REMOTE_JOBS_DIR
from drm4g                 import ( 
                                  DRM4G_DIR,
                                  DRM4G_RESOURCES_CONF,
                                  COMMUNICATORS,
                                  RESOURCE_MANAGERS,
                                  )

try :
    import configparser
except ImportError :
    import ConfigParser as configparser


logger = logging.getLogger(__name__)
pickled_file = os.path.join(DRM4G_DIR, "var", "rocci_pickled")

class ConfigureException(Exception):
    pass

class Configuration(object):
    """
    Configuration class provides facilities to:

    * parse DRM4G_RESOURCES_CONF resources
    * check key resources
    * instantiate objects such as communicators or managers

    """
    def __init__(self):
        self.resources  = dict()
        if not os.path.exists( DRM4G_RESOURCES_CONF ):
            assert DRM4G_RESOURCES_CONF, "resources.conf does not exist, please provide one"
        self.init_time = os.stat( DRM4G_RESOURCES_CONF ).st_mtime

    def check_update(self):
        """
        It checks if DRM4G file configuration has been updated.
        """
        if os.stat(DRM4G_RESOURCES_CONF).st_mtime != self.init_time:
            self.init_time = os.stat(DRM4G_RESOURCES_CONF).st_mtime
            return True
        else:
            return False

    def load(self):
        """
        Read the configuration file.
        """
        logger.debug("Reading file '%s' ..." % DRM4G_RESOURCES_CONF)
        try:
            try:
                conf_file   = open(DRM4G_RESOURCES_CONF, 'r')
                parser = configparser.RawConfigParser()
                try:
                    parser.readfp( conf_file , DRM4G_RESOURCES_CONF )
                except Exception as err:
                    output = "Configuration file '%s' is unreadable or malformed: %s" % ( DRM4G_RESOURCES_CONF , str( err ) )
                    logger.error( output )

                for sectname in parser.sections():
                    name                   = sectname
                    logger.debug(" Reading configuration for resource '%s'." % name )
                    self.resources[ name ] = dict( parser.items( sectname ) )

                    if self.resources[ name ][ 'lrms' ] == "rocci" :
                        if os.path.exists( os.path.join( pickled_file+"_"+name ) ):
                            try:
                                instances = []
                                with open( pickled_file+"_"+name, "r" ) as pf :
                                    while True :
                                        try:
                                            instances.append( pickle.load( pf ) )
                                        except EOFError :
                                            break
                                if not instances :
                                    pass
                                if instances:
                                    for instance in instances :
                                        insdict = dict()
                                        insdict['username'] = instance.vm_user
                                        insdict['frontend'] = instance.ext_ip
                                        insdict['communicator'] = instance.vm_comm
                                        insdict['private_key'] = instance.private_key
                                        insdict['enable'] = 'true'
                                        insdict['lrms'] = 'fork'
                                        insdict['max_jobs_running'] = instance.max_jobs_running
                                        self.resources[ name+"_"+instance.ext_ip ] = insdict
                                        logger.debug("Resource '%s' defined by: %s.",
                                                name+"_"+instance.ext_ip, ', '.join([("%s=%s" % (k,v)) for k,v in sorted(self.resources[name+"_"+instance.ext_ip].items())]))
                            except Exception as err :
                                raise Exception( "Could not create resource for the VMs of %s:\n%s" % (name,str(err)) )

                    logger.debug("Resource '%s' defined by: %s.",
                             sectname, ', '.join([("%s=%s" % (k,v)) for k,v in sorted(self.resources[name].items())]))
            except Exception as err:
                output = "Error reading '%s' file: %s" % (DRM4G_RESOURCES_CONF, str(err))
                logger.error( output )
        finally:
            conf_file.close()

    def check(self):
        """
        Check if the drm4g.conf file has been configured well.

        Return a list with the errors.
        """
        errors = []
        for resname, resdict in list(self.resources.items()) :
            logger.debug("Checking resource '%s' ..." % resname)

            reslist = list(resdict.keys( ))
            for key in reslist:
                if key not in ['enable', 'communicator', 'username', 'frontend', 'private_key', 'public_key', 'scratch', 'lrms', 'queue', 'max_jobs_in_queue', 'max_jobs_running', 'parallel_env', 'project', 'vo', 'host_filter', 'bdii', 'myproxy_server', 'vm_user', 'vm_communicator', 'vm_config', 'cloud_provider', 'flavour', 'virtual_image', 'instances', 'volume'] :
                    output = "'%s' resource has an invalid key : '%s'" % (resname, key)
                    logger.error( output )
                    errors.append( output )
            for key in [ 'enable' , 'frontend' , 'lrms' , 'communicator' ] :
                if not key in reslist :
                    output = "'%s' resource does not have '%s' key" % (resname, key)
                    logger.error( output )
                    errors.append( output )
            if ( resdict[ 'lrms' ] == 'rocci' and not resdict.get( 'private_key' ) ) :
                output = "'private_key' key has not been defined for '%s' resource" % resname
                logger.error( output )
                errors.append( output )
            if ( not 'max_jobs_running' in reslist ) and ( resdict[ 'lrms' ] != 'cream' ) :
                output = "'max_jobs_running' key is mandatory for '%s' resource" % resname
                logger.error( output )
                errors.append( output )
            if ( resdict[ 'lrms' ] != 'cream' and not resdict.get( 'max_jobs_running' ) ) :
                output = "'max_jobs_running' key has a wrong value for '%s' resource" % resname
                logger.error( output )
                errors.append( output )
            if ( not ( 'max_jobs_in_queue' in reslist ) and ( 'max_jobs_running' in reslist ) and ( resdict[ 'lrms' ] != 'cream' ) ) :
                self.resources[resname]['max_jobs_in_queue'] = resdict['max_jobs_running']
                logger.debug( "'max_jobs_in_queue' will be the same as the 'max_jobs_running'" )
            if ( not 'queue' in reslist ) and ( resdict[ 'lrms' ] != 'cream' ) :
                self.resources[resname]['queue'] = "default"
                output = "'queue' key will be called 'default' for '%s' resource" % resname
                logger.debug( output )
            if 'max_jobs_running' in reslist and resdict[ 'lrms' ] != 'cream' and resdict.get( 'max_jobs_in_queue' ).count( ',' ) !=  resdict.get( 'queue' ).count( ',' ) :
                output = "The number of elements in 'max_jobs_in_queue' are different to the elements of 'queue'"
                logger.error( output )
                errors.append( output )
            if 'max_jobs_running' in reslist and resdict[ 'lrms' ] != 'cream' and resdict.get( 'max_jobs_running' ).count( ',' ) !=  resdict.get( 'queue' ).count( ',' ) :
                output = "The number of elements in 'max_jobs_running' are different to the elements of 'queue'"
                logger.error( output )
                errors.append( output )
            if resdict[ 'lrms' ] != 'cream' and ( 'host_filter' in reslist ) :
                output = "'host_filter' key is only available for 'cream' lrms"
                logger.error( output )
                errors.append( output )
            if not self._exist_manager(COMMUNICATORS, resdict[ 'communicator' ] ):
                output = "'%s' has a wrong communicator: '%s'" % (resname , resdict[ 'communicator' ] )
                logger.error( output )
                errors.append( output )
            if resdict[ 'communicator' ] != 'local' and 'username' not in resdict :
                output = "'username' key is mandatory for '%s' communicator, '%s' resource" % (resdict[ 'communicator' ], resname)
                logger.error( output )
                errors.append( output )
            if not self._exist_manager(RESOURCE_MANAGERS, resdict[ 'lrms' ] ):
                output = "'%s' has a wrong lrms: '%s'" % ( resname , resdict[ 'lrms' ] )
                logger.error( output )
                errors.append( output )
            if resdict[ 'communicator' ] != 'local' :
                private_key = resdict.get( 'private_key' )
                if not private_key :
                    output = "'private_key' key is mandatory for '%s' resource" % resname
                    logger.error( output )
                    errors.append( output )
                else :
                    abs_private_key = os.path.expandvars( os.path.expanduser( private_key ) )
                    if not os.path.isfile( abs_private_key ) :
                        output = "'%s' does not exist for '%s' resource" % ( private_key , resname )
                        logger.error( output )
                        errors.append( output )
                    else :
                        self.resources[resname]['private_key'] = abs_private_key
                    public_key = resdict.get( 'public_key' )
                    if not public_key :
                        abs_public_key = abs_private_key + '.pub'
                    else :
                        abs_public_key = os.path.expandvars( os.path.expanduser( public_key ) )
                    if not os.path.isfile( abs_private_key ) :
                        output = "'%s' does not exist for '%s' resource" % ( abs_public_key , resname )
                        logger.error( output )
                        errors.append( output )
                    else :
                        self.resources[resname]['public_key'] = abs_public_key
            grid_cert = resdict.get( 'grid_cert' )
            if grid_cert :
                abs_grid_cert = os.path.expandvars( os.path.expanduser( grid_cert ) )
                if not os.path.isfile( abs_grid_cert ) :
                    output = "'%s' does not exist for '%s' resource" % ( abs_grid_cert , resname )
                    logger.error( output )
                    errors.append( output )
                else :
                    self.resources[resname]['grid_cert'] = abs_grid_cert
        '''
        #This is quite redundant
        #When running "drm4g resource edit", an exception appears if you have an error saying
        #"Please, review your configuration file", that is thrown by drm4g/commands
        if errors:
            output="Modify your configuration file before trying again."
            logger.error( output )
        '''
        return errors

    def make_communicators(self):
        """
        Make communicator objects corresponding to the configured resources.

        Return a dictionary, mapping the resource name into the corresponding objects.
        """
        communicators = dict()
        for name, resdict in list(self.resources.items()):
            try:
                communicator              = self._import_manager(COMMUNICATORS, resdict[ 'communicator' ] )
                com_object                = getattr( communicator , 'Communicator' ) ()
                com_object.username       = resdict.get( 'username' )
                com_object.frontend       = resdict.get( 'frontend' )
                com_object.private_key    = resdict.get( 'private_key' )
                com_object.public_key     = resdict.get( 'public_key' )
                com_object.work_directory = resdict.get( 'scratch', REMOTE_JOBS_DIR )
                communicators[name]       = com_object
                logger.debug("Communicator of resource '%s' is defined by: %s.",
                    name, ', '.join([("%s=%s" % (k,v)) for k,v in sorted(communicators[name].__dict__.items())]))
            except Exception as err:
                output = "Failed creating communicator for resource '%s' : %s" % ( name, str( err ) )
                logger.warning( output , exc_info=1 )
        return communicators

    def make_resources(self):
        """
        Make manager objects corresponding to the configured resources.

        Return a dictionary, mapping the resource name into the corresponding objects.
        """
        resources = dict()
        for name, resdict in list(self.resources.items()):
            try:
                resources[name]             = dict()
                manager                     = self._import_manager(RESOURCE_MANAGERS, resdict[ 'lrms' ] )
                resource_object             = getattr( manager , 'Resource' ) ()
                resource_object.name        = name
                resource_object.features    = resdict
                job_object                  = getattr( manager , 'Job' ) ()
                job_object.resfeatures      = resdict
                resources[name]['Resource'] = resource_object
                resources[name]['Job']      = job_object
            except Exception as err:
                output = "Failed creating objects for resource '%s' of type : %s" % ( name, str( err ) )
                logger.warning( output , exc_info=1 )
        return resources
    
    def _import_manager(self, MANAGERS, manager):
        """
        Auxiliar function to return module imported from MANAGERS, been referencied by manager.
        """
        if manager in MANAGERS :
            output = "Importing manager '%s' from '%s'" % ( manager , MANAGERS[ manager ] )
            logger.debug( output )
            return import_module(MANAGERS[ manager ] )
        else :
            output = "Importing manager '%s'" % ( manager )
            logger.debug( output )
            return import_module( manager )
    
    def _exist_manager(self, MANAGERS, manager):
        try:
            self._import_manager(MANAGERS, manager)
        except ImportError:
            return False
        else:
            return True

        



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import drm4g.managers
from string import Template

import logging
logger  = logging.getLogger(__name__)

# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
PBSNODES = 'LANG=POSIX pbsnodes' #pbsnodes - pbs node manipulation
QSUB     = 'LANG=POSIX qsub'     #qsub - submit pbs job
QSTAT    = 'LANG=POSIX qstat'    #qstat - show status of pbs batch jobs
QDEL     = 'LANG=POSIX qdel'     #qdel - delete pbs batch job

class Resource (drm4g.managers.Resource):

    def additional_queue_properties(self, queue):
        out, err = self.Communicator.execCommand('%s -q %s' % (QSTAT, queue.Name))
        #output line --> Queue Memory CPU_Time Walltime Node Run Que Lm State
        try:
            queueName, _, cpuTime, wallTime, _, _, _, lm = out.splitlines()[5].split()[0:8]
        except:
            logger.warning("Error parsing qstat: %s " % out)
        else:
            reTime = re.compile(r'(\d+):(\d+):\d+')
            if cpuTime != '--':
                try:
                    hours, minutes   = reTime.search(cpuTime).groups()
                    queue.MaxCpuTime = str(int(hours) * 60 + int(minutes))
                except:
                    pass
            if wallTime != '--':
                try:
                    hours, minutes   = reTime.search(wallTime).groups()
                    queue.MaxTime    = str(int(hours) * 60 + int(minutes))
                except:
                    pass
            if lm != '--':
                try:
                    queue.MaxRunningJobs = lm
                except:
                    pass
        return queue

class Job (drm4g.managers.Job):

    #pbs job status <--> GridWay job status
    states_pbs = {'E': 'ACTIVE',    #Job is exiting after having run.
                  'H': 'SUSPENDED', #Job is held.
                  'Q': 'PENDING',   #Job is queued, eligable to run or routed.
                  'R': 'ACTIVE',    #Job is running.
                  'T': 'PENDING',   #Job is being moved to new location.
                  'W': 'PENDING',   #Job is waiting for its execution time to be reached.
                  'S': 'SUSPENDED', #Job is suspend.
                  'C': 'DONE',	    #Job finalize.
                }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (QSUB, pathScript))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))
        return out.strip() #job_id

    def jobStatus(self):
        out, err = self.Communicator.execCommand('%s %s' % (QSTAT, self.JobId))
        if 'Unknown Job Id' in err :
            return 'DONE'
        elif err:
            return 'UNKNOWN'
        else:
            state = out.split()[-2]
            return self.states_pbs.setdefault(state, 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (QDEL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#PBS -N JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        if 'project' in parameters :
            args += '#PBS -P $project\n'
        if parameters['queue'] != 'default':
            args += '#PBS -q $queue\n'
        args += '#PBS -o $stdout\n'
        args += '#PBS -e $stderr\n'
        if 'maxWallTime' in parameters :
            args += '#PBS -l walltime=$maxWallTime\n'
        if 'maxCpuTime' in parameters :
            args += '#PBS -l cput=$maxCpuTime\n'
        if 'maxMemory' in parameters :
            args += '#PBS -l vmem=$maxMemoryMB\n'
        if 'ppn' in parameters and 'nodes' in parameters :
            args += '#PBS -l nodes=$nodes:ppn=$ppn\n'
        elif 'ppn' in parameters :
            node_count = int(parameters['count']) / int(parameters['ppn'])
            if node_count == 0:
                node_count = 1
            args += '#PBS -l nodes=%d:ppn=$ppn\n' % (node_count)
        else:
            args += '#PBS -l nodes=$count\n'
        args += '#PBS -v %s\n' % (','.join(['%s=%s' %(k, v) for k, v in list(parameters['environment'].items())]))
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import xml.dom.minidom


class Rsl2Parser(object):
    """
     Parser of RSL2 files. It is used like this:

             rsl    = RSL2Parser(filename)
             stdout = rsl.parse()
    """
    elems_to_parse = ['executable','stdout','stderr','directory','count','jobType',
                        'queue','maxTime','maxWallTime','maxCpuTime','minMemory','maxMemory','ppn', 'nodes']

    def __init__(self, filename):
        """
        @param filename : file name or file
        @type filename : string
        """
        self._values = { }
        self._str = xml.dom.minidom.parse(filename)

    def parseValue(self, name):
        """
        Gets data from <job><name></name></job> and adds it to self._values
        @param name : name of value
        @type name : string
        """
        val = self._str.getElementsByTagName(name)
        if val:
            self._values[name] = val[0].firstChild.data

    def parseEnvironment(self):
        """
        Gets all the data enclosed in <job><environment>...</environment></job> and adds
        it to self._values
        """
        environments = self._str.getElementsByTagName('environment')
        if environments:
            self._values['environment'] = dict((elem.getElementsByTagName('name')[0].firstChild.data,
                                           elem.getElementsByTagName('value')[0].firstChild.data) \
                                           for elem in environments)

    def parser(self):
        """
        Does the parsing. It is necessary to call this method after creating the object
        @return  : dictionary with RSL values
        """
        [self.parseValue(name) for name in self.elems_to_parse]
        self.parseEnvironment()
        return self._values




#!/usr/bin/env python

# ------------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------


import exceptions
import traceback
import sys
import tm_nordugrid

while True:
	try:
		tm_nordugrid.stop()
		reload(tm_nordugrid)
		tm_nordugrid.run()
	except exceptions.KeyboardInterrupt, e:
		sys.exit(-1)
	except exceptions.SystemExit, e:
		print e
		sys.exit(0)
	except Exception, e:
		print "Pero que muuuvi!!!"
		print e.__class__, e
		traceback.print_exc(file=sys.stdout)
	
	
	







#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the 
# "Licence"); 
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

from setuptools import setup
from setuptools import find_packages
from setuptools.command.install import install
from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop
import os
import subprocess
import glob
import sys
import stat
from pprint import pprint

#To ensure a script runs with a minimal version requirement of the Python interpreter
#assert sys.version_info >= (2,5)
if sys.version_info<(3,5):
  exit( 'The version number of Python has to be >= 3.5' )

here = os.path.abspath(os.path.dirname(__file__))
gridway_src = "gridway-5.8"
  
# read the contents of your README file
with open(os.path.join(here, 'README'), encoding='utf-8') as f:
  long_description = f.read()

MAKE_CLEAN = False
def build():
  current_path = os.getcwd()
  if not os.path.exists(gridway_src) :
      raise Exception("The specified directory %s doesn't exist" % gridway_src)
  
  #setting same mtime to gridway sources
  mtime_ref = os.path.getmtime(gridway_src)
  for root, dirs, files in os.walk(gridway_src):
    for name in files + dirs:
      os.utime(os.path.join(root, name), (mtime_ref, mtime_ref))
  os.utime(gridway_src, (mtime_ref, mtime_ref))
  #

  os.chdir( gridway_src )
  #to avoid re-run configure each time.
  if(not os.path.isfile('config.log') or os.path.getmtime('config.log') <= os.path.getmtime('configure') ):
    st = os.stat('configure')
    os.chmod('configure', st.st_mode | stat.S_IEXEC)
    os.makedirs('build',exist_ok=True)
    buildPath = os.path.abspath('build')
    exit_code = subprocess.call('./configure --prefix="%s"' % buildPath, shell=True)
    if exit_code:
      raise Exception("Configure failed - check config.log for more detailed information")
  
  exit_code = subprocess.call('make', shell=True)
  if exit_code:
    raise Exception("make failed")
  exit_code = subprocess.call('make install', shell=True)
  if exit_code:
    raise Exception("make install failed")
  if MAKE_CLEAN:
    exit_code = subprocess.call('make clean', shell=True)
    if exit_code:
      raise Exception("make clean failed")
  os.chdir( current_path )

gw_files = ('bin',
    [
      gridway_src + '/build/bin/gwuser',
      gridway_src + '/build/bin/gwacct',
      gridway_src + '/build/bin/gwwait',
      gridway_src + '/build/bin/gwhost',
      gridway_src + '/build/bin/gwhistory',
      gridway_src + '/build/bin/gwsubmit',
      gridway_src + '/build/bin/gwps',
      gridway_src + '/build/bin/gwkill',
      gridway_src + '/build/bin/gwd',
      gridway_src + '/build/bin/gw_flood_scheduler',
      gridway_src + '/build/bin/gw_sched',
    ])

#OLD WAY
#gw_files = ('bin',
#    [
#      gridway_src + '/src/cmds/gwuser',
#      gridway_src + '/src/cmds/gwacct',
#      gridway_src + '/src/cmds/gwwait',
#      gridway_src + '/src/cmds/gwhost',
#      gridway_src + '/src/cmds/gwhistory',
#      gridway_src + '/src/cmds/gwsubmit',
#      gridway_src + '/src/cmds/gwps',
#      gridway_src + '/src/cmds/gwkill',
#      gridway_src + '/src/gwd/gwd',
#      gridway_src + '/src/scheduler/gw_flood_scheduler',
#      gridway_src + '/src/scheduler/gw_sched',
#    ])

#pprint(vars(self))
class build_ext_wrapper(build_ext):
  def run(self):
    print("[running build_ext_wrapper.run() ...]")
    if(self.verbose):
      pprint(vars(self))
    build()
    build_ext.run(self)

class install_wrapper(install):
  def run(self):
    print("[running install_wrapper.run() ...]")
    if(self.verbose):
      pprint(vars(self))
    build()
    install.run(self)

class develop_wrapper(develop):
  def run(self):
    print("[running develop_wrapper.run() ...]")
    develop.run(self)
    if(self.verbose):
      pprint(vars(self))
    for filename in gw_files[1]:
      dst = os.path.join(self.script_dir, os.path.basename(filename))
      if(os.path.lexists(dst)):
        if(self.verbose):
          print("[removing %s]" % dst)
        os.remove(dst)
      src = os.path.abspath(filename)
      if(not self.uninstall):
        if(self.verbose):
          print("[creating symlink: %s -> %s]" % (dst, src))
        os.symlink(src,dst)


bin_scripts = glob.glob(os.path.join('bin', '*'))
#bin_scripts.append('LICENSE')

# FROM: https://github.com/jbweston/miniver
def get_version_and_cmdclass(package_name):
    import os
    from importlib.util import module_from_spec, spec_from_file_location
    spec = spec_from_file_location('version',
                                   os.path.join(package_name, '_version.py'))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass
version, cmdclass = get_version_and_cmdclass('drm4g')

setup(
  name='drm4g',
  version=version,
  python_requires=">=3.5",
  packages=find_packages(),
  include_package_data=True,
  package_data={'drm4g' : ['conf/*.conf', 'conf/job_template.default', 'conf/*.sh']},
  data_files = [gw_files],
  author='Santander Meteorology Group (UC-CSIC)',
  author_email='antonio.cofino@unican.es',
  url='https://github.com/SantanderMetGroup/DRM4G',
  project_urls = {
    'Documentation' : 'https://meteo.unican.es/trac/wiki/DRM4G'           ,
    'Source'        : 'https://github.com/SantanderMetGroup/DRM4G'        ,
    'Tracker'       : 'https://github.com/SantanderMetGroup/DRM4G/issues' ,
    'Download'      : 'https://pypi.org/project/drm4g/#files'             , 
    'Twitter'       : 'https://twitter.com/SantanderMeteo'

  },
  license='European Union Public License 1.1',
  description='Meta-scheduling framework for distributed computing infrastructures',
  long_description=long_description,
  long_description_content_type='text/markdown',
  classifiers=[
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
    "Operating System :: POSIX",
    "Topic :: Scientific/Engineering",
    "Topic :: Office/Business :: Scheduling",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
  ],
  install_requires=['fabric', 'docopt', 'openssh-wrapper','scp'],
  scripts=bin_scripts,
  cmdclass={
    'build_ext' : build_ext_wrapper,
    'install'   : install_wrapper,
    'develop'   : develop_wrapper,
    'sdist'     : cmdclass['sdist'],
    'build_py'  : cmdclass['build_py'],
  },
)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import drm4g.managers
from string import Template

import logging
logger  = logging.getLogger(__name__)


SH = '/bin/bash'

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (SH, pathScript))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))
        job_id = out.rstrip('\n')
        return job_id

    def jobStatus(self):
        out, err = self.Communicator.execCommand('ps --no-heading -p %s' %(self.JobId))
        if out:
            return 'ACTIVE'
        else:
            return 'DONE'

    def jobCancel(self):
        jobs_to_kill = [self.JobId]
        while jobs_to_kill:
            for job in jobs_to_kill:
                out, err = self.Communicator.execCommand('ps ho pid --ppid %s' % (job))
                jobs_to_kill = [line.lstrip() for line in out.splitlines()] + jobs_to_kill
                out, err = self.Communicator.execCommand('kill -9 %s' % (job))
                if err:
                    raise drm4g.managers.JobException('Could not kill %s : %s' % (job, ' '.join(err.split('\n'))))
                jobs_to_kill.remove(job)

    def jobTemplate(self, parameters):
        line  = '#!/bin/bash\n'
        line += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        line += '\n'
        line += 'nohup $executable 2> $stderr > $stdout &\n'
        line += 'echo $$!\n'
        return Template(line).safe_substitute(parameters)





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import os
from os.path     import join, expanduser, exists, basename

import io
import re
import time
import signal
import traceback
import threading
import subprocess
import drm4g.communicators
import drm4g.commands
from drm4g.commands         import Agent
from drm4g.communicators    import SSH_CONNECT_TIMEOUT, SFTP_CONNECTIONS
from drm4g                  import DRM4G_DIR, DRM4G_DIR_VAR 
from drm4g.utils.url        import urlparse
from openssh_wrapper import SSHConnection

import logging
logger  = logging.getLogger(__name__)


class Communicator(drm4g.communicators.Communicator):
    """
    Create a SSH session to remote resources.
    """
    _lock       = __import__('threading').Lock()
    _sem        = __import__('threading').Semaphore(SFTP_CONNECTIONS)
    _trans      = None

    socket_dir=None

    def __init__(self):
        super(Communicator,self).__init__()
        #logger.debug("\n\nCREATING NEW COMMUNICATOR\n%s\n" % traceback.format_exc())
        self.conn=None
        self.parent_module=None
        self.configfile=None

        # module_dict = {'rocci.py':'rocci', 'im_mad.py':'im', 'tm_mad.py':'tm', 'em_mad.py':'em'}
        # #t=repr(traceback.format_stack())
        # trace = traceback.extract_stack()
        # for module_path in trace :
        #     module = basename(module_path[0])
        #     if module in module_dict :
        #         self.parent_module = module_dict[module]
        #         self.configfile = join(DRM4G_DIR, 'etc', 'openssh_%s.conf' % module_dict[module])
        #         break
        self.get_parent_module()
        
        with self._lock:
            self.agent=Agent()
            self.agent.start()
        self.agent_socket=self.agent.update_agent_env()['SSH_AUTH_SOCK']

        if not Communicator.socket_dir:
            Communicator.socket_dir=join(DRM4G_DIR_VAR, 'sockets')

    def get_parent_module(self):
        module_dict = {'rocci.py':'rocci', 'im_mad.py':'im', 'tm_mad.py':'tm', 'em_mad.py':'em'}
        #t=repr(traceback.format_stack())
        trace = traceback.extract_stack()
        for module_path in trace :
            module = basename(module_path[0])
            if module in module_dict :
                self.parent_module = module_dict[module]
                self.configfile = join(DRM4G_DIR, 'etc', 'openssh_%s.conf' % self.parent_module)
                break
            else:
                #if it's none of the other modules but it's still calling Configuration().make_communicators
                #it has to be from commands/__init__.py's check_frontends() being called from commands/resource.py 
                #for which we had decided the 'im' socket would be used
                self.parent_module = 'im' 
                self.configfile = join(DRM4G_DIR, 'etc', 'openssh_%s.conf' % self.parent_module)

    def createConfFiles(self):
        logger.debug("Running createConfFiles function from %s" % self.parent_module)
        #the maximum length of the path of a unix domain socket is 108 on Linux, 104 on Mac OS X
        conf_text = ("Host *\n"
            "    ControlMaster auto\n"
            "    ControlPath %s/%s-%s\n"
            "    ControlPersist 10m\n"
            "    StrictHostKeyChecking no\n"
            "    LogLevel ERROR")

        for manager in ['im', 'tm', 'em', 'rocci']:
            with io.FileIO(join(DRM4G_DIR, 'etc', 'openssh_%s.conf' % manager), 'w') as conf_file:
                conf_file.write(conf_text % (Communicator.socket_dir, manager, '%r@%h:%p'))
        try:
            if not exists(Communicator.socket_dir):
                logger.debug("Creating socket directory in %s" % Communicator.socket_dir)
                os.makedirs(Communicator.socket_dir)
        except OSError as excep:
            if "File exists" in str(excep):
                logger.warning("The directory %s already exists" % Communicator.socket_dir)
            else:
                logger.error("An unexpected exception ocurred:\n"+str(excep))
        logger.debug("Ending createConfFiles function from %s" % self.parent_module)

    def connect(self):
        """
        To establish the connection to resource.
        """
        try:
            p_module=sys._getframe().f_back.f_code.co_filename
            p_function=sys._getframe().f_back.f_code.co_name
            t=repr(traceback.format_stack())
            #for line in t:
            #    print line

            #logger.debug("Running connect function\n    - called from "+p_module+"\n    - by function "+p_function)
            logger.debug("\n\nRunning connect function from %s\n    - called from %s\n    - by function %s\nTraceback :\n%s\n" % (self.parent_module, p_module, p_function, t))
            #logger.debug("\np_module = %s\np_function = %s\n" % (p_module, p_function))

            if not self.configfile:
                logger.debug("Variable 'self.configfile' is not defined")
                logger.error("OpenSHH configuration file's path is not defined")

            if not exists(self.configfile):
                self.createConfFiles()
            logger.debug("The socket "+join(Communicator.socket_dir, '%s-%s@%s:%s' % (self.parent_module ,self.username, self.frontend, self.port))+" existance is "+str(exists(join(Communicator.socket_dir, '%s-%s@%s:%s' % (self.parent_module ,self.username, self.frontend, self.port)))))

            if not exists(join(Communicator.socket_dir, '%s-%s@%s:%s' % (self.parent_module ,self.username, self.frontend, self.port))):
                logger.debug("No master conncection exists for %s so a new one will be created" % self.parent_module)
                def first_ssh():
                    try:
                        logger.debug("Running first_ssh function\n    - Creating first connection for %s" % self.parent_module)
                        #this is here because the threads are created at the same time, so the moment one creates the conection, the rest are going to cause an UnboundLocalError exception
                        #(which probably shouldn't be ocurring since ControlMaster is set to auto - only if they execute this at the same time)
                        if not exists(join(Communicator.socket_dir, '%s-%s@%s:%s' % (self.parent_module ,self.username, self.frontend, self.port))):
                            command = 'ssh -F %s -i %s -p %s -T %s@%s' % (self.configfile, self.private_key, str(self.port), self.username, self.frontend)
                        pipe = subprocess.Popen(command.split(), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        out,err = pipe.communicate()

                        if err:
                            if "too long for Unix domain socket" in str(err) or "ControlPath too long" in str(err):
                                logger.debug("Socket path was too long for Unix domain socket.\n    Creating sockets in ~/.ssh/dmr4g.\n    Exception captured in first_ssh.")
                                self._change_socket_dir()
                                logger.debug("Calling first_ssh once again, but with a new socket_dir")
                                first_ssh()
                            elif "disabling multiplexing" in str(err):
                                logger.debug("connect function: The multiplexing of connections isn't working. Eliminating "+self.parent_module+"'s socket file.")
                                self._delete_socket()
                                first_ssh()
                            elif "bind: No such file or directory" in str(err) or "cannot bind to path" in str(err):
                                logger.debug("The connection through the socket %s-%s@%s:%s wasn't established since the socket directory %s hasn't been created yet." % (self.parent_module ,self.username, self.frontend, self.port, Communicator.socket_dir))
                                self.createConfFiles()
                                first_ssh()
                            else:
                                logger.debug("Unexpected error occured while running first_ssh:\n"+str(err))
                                logger.warning(str(err))
                    except UnboundLocalError as err:
                        logger.warning( "Local variable referenced before assignment" )
                        logger.debug( str(err) )

                t = threading.Thread(target=first_ssh, args = ())
                t.daemon = True
                logger.debug("Starting thread with first_ssh")
                t.start()
                time.sleep(5) #so that there's time to make the first connection in case there was an error

            if self.conn==None:
                logger.debug("No conn exists (conn == "+str(self.conn)+") for "+self.parent_module+" so a new one will be created.")
                self.conn = SSHConnection(self.frontend, login=self.username, port=str(self.port),
                    configfile=self.configfile, identity_file=self.private_key,
                    ssh_agent_socket=self.agent_socket, timeout=SSH_CONNECT_TIMEOUT)

            logger.debug("Ending connect function from %s" % self.parent_module)

        except Exception as excep :
            if "too long for Unix domain socket" in str(excep):
                logger.debug("Socket path was too long for Unix domain socket.\n    Creating sockets in ~/.ssh/dmr4g.\n    Exception captured in connect's except.")
                self._change_socket_dir()
                self.connect()
            else:
                logger.error(str(excep))
                raise

    def execCommand(self , command , input = None ):
        try:
            logger.debug("Running execCommand function from "+self.parent_module+"\n    - Trying to execute command "+str(command))

            if not self.conn:
                logger.debug("Going to run connect function.\n    - That should already have been done, so it shouldn't do anything.")
                self.connect()

            ret = self.conn.run(command)
            logger.debug("Ending execCommand function.")
            return ret.stdout , ret.stderr
        except Exception as excep:
            if "disabling multiplexing" in str(excep):
                logger.debug("Mux isn't working from the execCommand function. Eliminating "+self.parent_module+"'s socket file.")
                self._delete_socket()
                self.execCommand(command, input)
            else:
                logger.warning(str(excep))

    def mkDirectory(self, url):
        try:
            logger.debug("Running mkDirectory function from %s" % self.parent_module)
            to_dir         = self._set_dir(urlparse(url).path)
            stdout, stderr = self.execCommand( "mkdir -p %s" % to_dir )
            if stderr :
                logger.warning( "Could not create %s directory: %s" % ( to_dir , stderr ) )
            logger.debug("Ending mkDirectory function from %s" % self.parent_module)
        except Exception as excep:
            if "disabling multiplexing" in str(excep):
                logger.debug("Mux isn't working from the mkDirectory function. Eliminating "+self.parent_module+"'s socket file.")
                self._delete_socket()
                self.mkDirectory(url)
            else:
                logger.warning(str(excep))

    def rmDirectory(self, url):
        try:
            logger.debug("Running rmDirectory function from %s" % self.parent_module)
            to_dir         = self._set_dir(urlparse(url).path)
            stdout, stderr = self.execCommand( "rm -rf %s" % to_dir )
            if stderr:
                logger.warning( "Could not remove %s directory: %s" % ( to_dir , stderr ) )
            logger.debug("Ending rmDirectory function from %s" % self.parent_module)
        except Exception as excep:
            if "disabling multiplexing" in str(excep):
                logger.debug("Mux isn't working from the rmDirectory function. Eliminating "+self.parent_module+"'s socket file.")
                self._delete_socket()
                self.rmDirectory(url)
            else:
                logger.warning(str(excep))

    def copy( self , source_url , destination_url , execution_mode = '' ) :
        try:
            logger.debug("Running copy function from %s" % self.parent_module)
            if not self.conn:
                self.connect()
            with self._sem :
                if 'file://' in source_url :
                    from_dir = urlparse( source_url ).path
                    to_dir   = self._set_dir( urlparse( destination_url ).path )
                    self.conn.scp( [from_dir] , target=to_dir )
                    if execution_mode == 'X':
                        stdout, stderr = self.execCommand( "chmod +x %s" % to_dir )
                        if stderr :
                            logger.warning( "Could not change access permissions of %s file: %s" % ( to_dir , stderr ) )
                else:
                    from_dir = self._set_dir( urlparse( source_url ).path )
                    to_dir   = urlparse(destination_url).path
                    self.remote_scp( [from_dir] , target=to_dir )
            logger.debug("Ending copy function from %s" % self.parent_module)
        except Exception as excep:
            if "disabling multiplexing" in str(excep):
                logger.debug("Mux isn't working from the copy function. Eliminating "+self.parent_module+"'s socket file.")
                self._delete_socket()
                self.copy(source_url , destination_url)
            else:
                logger.warning(str(excep))

    def _change_socket_dir(self):
        logger.debug("Running _change_socket_dir function from %s" % self.parent_module)
        try:
            if exists(Communicator.socket_dir):
                os.rmdir(Communicator.socket_dir)
        except OSError as excep:
            if "No such file or directory" in str(excep):
                logger.debug("The old socket directory %s has already been deleted" % Communicator.socket_dir)
            else:
                logger.warning(str(excep))
        Communicator.socket_dir = join(expanduser('~'), '.ssh/drm4g')
        self.createConfFiles()
        logger.debug("Ending _change_socket_dir function from %s" % self.parent_module)

    def _delete_socket(self):
        try:
            logger.debug("Running _delete_socket function from %s" % self.parent_module)
            os.remove("%s/%s-%s@%s:%s" % (Communicator.socket_dir,self.parent_module,self.username,self.frontend,str(self.port)))
            logger.debug("Ending _delete_socket function from %s" % self.parent_module)
        except OSError as excep:
            if "No such file or directory" in str(excep):
                logger.debug("The socket %s/%s-%s@%s:%s does not exist" % (Communicator.socket_dir,self.parent_module,self.username,self.frontend,str(self.port)))
                logger.debug("Ending _delete_socket function from %s" % self.parent_module)
            else:
                logger.warning(str(excep))

    #internal
    def _set_dir(self, path):
        logger.debug("Running _set_dir function from %s" % self.parent_module)
        work_directory =  re.compile( r'^~' ).sub( self.work_directory , path )
        logger.debug("Ending _set_dir function from %s" % self.parent_module)
        return  work_directory

    def remote_scp(self, files, target):
        logger.debug("Running remote_scp function from %s" % self.parent_module)
        scp_command = self.scp_command(files, target)
        pipe = subprocess.Popen(scp_command,
                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, env=self.get_env())

        signal.alarm(SSH_CONNECT_TIMEOUT)
        err = ''
        try:
            _, err = pipe.communicate()
        except IOError as exc:
            #pipe.terminate() # only in python 2.6 allowed
            os.kill(pipe.pid, signal.SIGTERM)
            signal.alarm(0)  # disable alarm
            logger.warning("%s (under %s): %s" % (' '.join(scp_command), self.username, str(exc)))
        signal.alarm(0)  # disable alarm
        returncode = pipe.returncode
        if returncode != 0:  # ssh client error
            logger.warning("%s (under %s): %s" % (' '.join(scp_command), self.username, err.strip()))
        logger.debug("Ending remote_scp function from %s" % self.parent_module)

    def scp_command(self, files, target, debug=False):
        """
        Build the command string to transfer the files identified by filenames.
        Include target(s) if specified. Internal function
        """
        logger.debug("Running scp_command function from %s" % self.parent_module)
        cmd = ['scp', debug and '-vvvv' or '-q', '-r']

        if self.username:
            remotename = '%s@%s' % (self.username, self.frontend)
        else:
            remotename = self.frontend
        if self.configfile:
            cmd += ['-F', self.configfile]
        if self.private_key:
            cmd += ['-i', self.private_key]
        if self.port:
            cmd += ['-P', str(self.port)]

        if not isinstance(files, list):
            logger.warning('"files" argument has to be an iterable (list or tuple)')
        if len(files) < 1:
            logger.warning('You should name at least one file to copy')

        for f in files:
            cmd.append('%s:%s' % (remotename, f))
        cmd.append(target)
        logger.debug("The command is "+str(cmd))
        logger.debug("Ending scp_command function from %s" % self.parent_module)
        return cmd

    def get_env(self):
        """
        Retrieve environment variables and replace SSH_AUTH_SOCK
        if ssh_agent_socket was specified on object creation.
        """
        logger.debug("Running get_env function from %s" % self.parent_module)
        env = os.environ.copy()
        if self.agent_socket:
            env['SSH_AUTH_SOCK'] = self.agent_socket
        logger.debug("Ending get_env function from %s" % self.parent_module)
        return env



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import sys
import re
import time
import signal
import getpass
import subprocess
import datetime
import pprint

from drm4g             import REMOTE_VOS_DIR, DRM4G_RESOURCES_CONF, DRM4G_DIR,console_logger, DRM4G_DIR_VAR
#from drm4g.managers    import rocci
from drm4g.core.im_mad import GwImMad
from os.path           import expanduser, join, dirname, exists, basename, expandvars


PY2 = sys.version_info[0] == 2


def process_is_runnig( pid ):
    """
    Check is a process is running given a file
    """
    try:
        with open( pid , 'r' ) as f:
            lines = f.readlines()
        os.kill( int( lines[0].strip() ) , 0 )
    except :
        return False
    else:
        return True

def exec_cmd( cmd , stdin=subprocess.PIPE, stdout=subprocess.PIPE,
              stderr=subprocess.STDOUT, env=os.environ ):
    """
    Execute shell commands
    """
    console_logger.debug( "Executing command ... " + cmd )
    cmd_to_exec = subprocess.Popen(  cmd ,
                                  shell=True ,
                                  stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  env=env
                                  )
    out , err =  cmd_to_exec.communicate()
    return out.decode() , err.decode()

def yes_no_choice( message ,  default = 'y' ) :
    """
    Ask for Yes/No questions
    """
    choices = 'Y/n' if default.lower() in ( 'y', 'yes' ) else 'y/N'
    if PY2 :
        choice = raw_input( "%s (%s) " % ( message, choices ) )
    else :
        choice = input( "%s (%s) " % ( message, choices ) )
    values = ( 'y', 'yes', '' ) if default == 'y' else ( 'y', 'yes' )
    return choice.strip().lower() in values

class Agent( object ):
    """
    Class to manage ssh-agent command.
    """

    def __init__( self, resource = None ):
        if resource :
            self.private_key  = resource[ 'private_key' ]
            self.public_key   = resource[ 'public_key' ]
            self.user         = resource[ 'username' ]
            self.frontend     = resource[ 'frontend' ]
        self.agent_env    = dict()
        self.agent_file   = join(DRM4G_DIR_VAR, 'agent.conf' )

    def start( self ):
        def _start():
            # 's' option generates Bourne shell commands on stdout
            out , err = exec_cmd( 'ssh-agent -s ' )
            console_logger.debug( out )
            match = re.search( 'SSH_AUTH_SOCK=(?P<SSH_AUTH_SOCK>[^;]+);.*' \
                           + 'SSH_AGENT_PID=(?P<SSH_AGENT_PID>\d+);', out, re.DOTALL)
            if match :
                console_logger.debug( "  OK" )
                self.agent_env = match.groupdict()
                console_logger.debug( '  Agent pid: %s'  % self.agent_env['SSH_AGENT_PID'])
            else:
                console_logger.error( err )
                raise Exception( '  Cannot determine agent data from output: %s' % out )
            with open( self.agent_file , 'w') as f:
                f.write( self.agent_env['SSH_AGENT_PID'] + '\n' + self.agent_env['SSH_AUTH_SOCK'] )
        console_logger.debug('Starting ssh-agent ...')
        if not self.is_alive() :
            _start()
        elif self.is_alive() :
            console_logger.debug( "  ssh-agent is already running" )
        elif not self.agent_env:
            self.get_agent_env()

    def status( self ) :
        if self.is_alive() :
            console_logger.info( "ssh-agent is running" )
        else :
            console_logger.info( "ssh-agent is stopped" )

    def is_alive( self ):
        if not exists( self.agent_file ) :
            console_logger.debug("  '%s' does not exist" % ( self.agent_file ) )
            return False
        else :
            if process_is_runnig( self.agent_file ):
                return True
            else :
                return False

    def get_agent_env( self ):
        console_logger.debug("Reading '%s' file" % ( self.agent_file ) )
        with open( self.agent_file , 'r' ) as f:
            lines = f.readlines()
        self.agent_env['SSH_AGENT_PID'] = lines[0].strip()
        self.agent_env['SSH_AUTH_SOCK'] = lines[1].strip()

    def update_agent_env( self ):
        env = os.environ
        if not self.agent_env :
            self.get_agent_env()
        env.update( self.agent_env )
        return env

    def add_key( self, lifetime ):
        console_logger.info("--> Add '%s' into ssh-agent for %s hours" % ( self.private_key , lifetime ) )
        out , err = exec_cmd( 'ssh-add -t %sh %s' % ( lifetime , self.private_key ),
                  stdin=sys.stdin, env=self.update_agent_env() )
        mo = re.compile(r'.* (\d*) .*').search( err )
        if mo :
            console_logger.info( " Lifetime set to " + str( datetime.timedelta( seconds=int( mo.group(1) ) ) ) )
        else :
            console_logger.info( err )

    def delete_key( self ):
        console_logger.info('--> Remove key %s' % self.private_key )
        out , err = exec_cmd( 'ssh-add -d %s' % self.private_key,
                              stdin=sys.stdin, stdout=sys.stdout, env=self.update_agent_env() )
        if err :
            console_logger.info( err )

    def copy_key( self ):
        console_logger.info("--> Copying '%s' to ~/.ssh/authorized_keys file on '%s'" % ( self.private_key, self.frontend ) )

        private_key_path = expanduser(self.private_key)
        public_key_path = expanduser(self.public_key)

        error_message = "Could not find the %s in %s.\n" \
                "Be sure to first generate the authentication keys with 'ssh_keygen'" \
                " and then specify it's path when defining your resource.\n" \
                "By default '~/.ssh/id_rsa' and '~/.ssh/id_rsa.pub' will be " \
                "considered to be your private and public keys respectively"

        if not exists(private_key_path):
            raise Exception(error_message % ('private key', private_key_path))

        if not public_key_path:
            public_key_path = private_key_path+'.pub'

        if not exists(public_key_path):
            raise Exception(error_message % ('public key', public_key_path))

        cmd='cat %s | ssh %s@%s "mkdir -p ~/.ssh;' \
        ' cat > ~/temp_pub_key && grep -q -f temp_pub_key ~/.ssh/authorized_keys || cat temp_pub_key >> ~/.ssh/authorized_keys &&' \
        ' chmod 600 ~/.ssh/authorized_keys && chmod 700 ~/.ssh && rm temp_pub_key"' % (public_key_path, self.user, self.frontend)

        #answer = subprocess.Popen(cmd.split(), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) #does not work
        answer = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out,err = answer.communicate()

        if err:
            if "grep" in err:
                pass
            else:
                console_logger.info( err )
                raise Exception(err)
        console_logger.info( "The copy of the public key %s has been successful" % public_key_path )

    def list_key( self ):
        console_logger.info("--> Display '%s' key" % self.private_key )
        out , err = exec_cmd( 'ssh-add -L' , env=self.update_agent_env() )
        match = re.search( '.*%s' % basename( self.private_key ) , out )
        if match :
            console_logger.info( match.group() )
        else :
            console_logger.info( " The private key '%s' is not available anymore" % self.private_key )

    def stop( self ):
        console_logger.info( 'Stopping ssh-agent ... ' )
        if self.is_alive():
            out , err = exec_cmd( 'ssh-agent -k' , env=self.update_agent_env() )
            if out :
                console_logger.debug( out )
            if err :
                console_logger.info( err )
            else :
                console_logger.info( "  OK" )
        else:
            console_logger.warn( '  WARNING: ssh-agent is already stopped' )
        try:
            os.remove( self.agent_file )
        except :
            pass


class Daemon( object ):

    def __init__( self ):
        self.gwd_pid  = join(DRM4G_DIR_VAR , 'gwd.pid' )

    def status( self ):
        if self.is_alive() :
            console_logger.info( "DRM4G is running" )
        else :
            console_logger.info( "DRM4G is stopped" )

    def is_alive( self ):
        if not exists( self.gwd_pid ) :
            return False
        else :
            if process_is_runnig( self.gwd_pid ) :
                return True
            else :
                return False

    def start( self ):
        console_logger.info( "Starting DRM4G .... " )
        if not exists( self.gwd_pid ) or ( exists( self.gwd_pid ) and not process_is_runnig( self.gwd_pid ) ) :
            lock = join(DRM4G_DIR_VAR , '/.lock' )
            if exists( lock ) :
                os.remove( lock )
            console_logger.debug( "Starting gwd .... " )
            out , err = exec_cmd( 'gwd' )
            if err :
                console_logger.info( err )
            if out :
                console_logger.info( out )
            if not err and not out :
                console_logger.info( "  OK" )
        else :
            console_logger.info( "  WARNING: DRM4G is already running" )

    def stop( self ):
        console_logger.info( "Stopping DRM4G .... " )
        console_logger.debug( "Stopping gwd .... " )
        if self.is_alive():
            with open(self.gwd_pid) as pid_file:
                pid = next(pid_file)
                processes_to_kill = [pid]
                while processes_to_kill:
                    for process in processes_to_kill:
                        try:
                            cmd = "ps ho pid --ppid %s" % (process)
                            out , err = exec_cmd( cmd )
                            processes_to_kill = [line.lstrip() for line in out.splitlines()] + processes_to_kill
                            processes_to_kill.remove(process)
                            os.kill( int(process), signal.SIGTERM )
                        except OSError as err:
                            if "No such process" in err:
                                continue
                            raise
            while self.is_alive() :
                time.sleep(1)
            try:
                os.remove(self.gwd_pid)
            except OSError:
                console_logger.error("  '%s' does not exist" % self.gwd_pid)
            console_logger.info( "  OK" )
        else:
            console_logger.info("  WARNING: daemon is already stopped")

    def clear( self ):
        yes_choise = yes_no_choice( "Do you want to continue clearing DRM4G? " )
        if yes_choise :
            console_logger.info( "Clearing DRM4G .... " )
            cmd = "%s -c" % ( "gwd" )
            out , err = exec_cmd( cmd )
            console_logger.debug( out )
            if err :
                console_logger.info( err )
            if out :
                console_logger.info( out )
            if not err and not out :
                console_logger.info( "  OK" )
        else :
            self.start()

class Resource( object ):

    def __init__( self , config ):
        self.config = config

    def create_vms(self):
        """
        Creates a virtual machine with the information given through the resources.conf and cloudsetup.json files
        """
        self.check( )
        for resname, resdict in self.config.resources.items():
            if resdict[ 'lrms' ] in ['rocci']:
                rocci.manage_instances('start', resname, resdict)

    def destroy_vms(self):
        """
        Destroy running virtual machines
        """
        self.check( )
        for resname, resdict in self.config.resources.items():
            if resdict[ 'lrms' ] in ['rocci']:
                rocci.manage_instances('stop', resname, resdict)

    def update_hosts(self):
        """
        Forces the host list to be updated
        """
        try:
            GwImMad().do_DISCOVER("discover - - -", False)
            #GwImMad().do_MONITOR("monitor 2 %s -" % resname)#, False)
        except Exception as err:
            console_logger.error( "Could not update hosts:\n%s" % str(err))

    def list_resources(self):
        """
        List all resources, the ones configured by the user and the ones configured internally
        For example, when creating cloud virtual machines
        """
        self.check( )
        console_logger.info("Resources:")
        for resname, resdict in self.config.resources.items():
            console_logger.info("    "+str(resname))
            console_logger.info("        communicator:  "+str(resdict['communicator']))
            if 'username' in resdict.keys():
                console_logger.info("        username:      "+str(resdict['username']))
            console_logger.info("        frontend:      "+str(resdict['frontend']))
            if 'private_key' in resdict.keys():
                console_logger.info("        private key:   "+str(resdict['private_key']))
            console_logger.info("        lrms:          "+str(resdict['lrms']))

    def check_frontends( self ) :
        """
        Check if the frontend of a given resource is reachable.
        """
        self.check( )
        communicators = self.config.make_communicators()
        for resname, resdict in sorted( self.config.resources.items() ) :
            if resdict[ 'enable' ] == 'true' :
                communicator = communicators.get( resname )
                try :
                    # if resdict[ 'communicator' ] == 'op_ssh' :
                    #     #it will use im's socket instead of creating a new one
                    #     communicator.parent_module = 'im'
                    #     communicator.configfile = join(DRM4G_DIR, 'etc', 'openssh_im.conf')
                    communicator.connect()
                    console_logger.info( "Resource '%s' :" % ( resname ) )
                    console_logger.info( "--> The front-end '%s' is accessible\n" % communicator.frontend )
                except Exception as err :
                    console_logger.error( "Resource '%s' :" % ( resname ) )
                    console_logger.error( "--> The front-end '%s' is not accessible\n" % communicator.frontend )

    def edit( self ) :
        """
        Edit resources file.
        """
        console_logger.debug( "Editing '%s' file" % DRM4G_RESOURCES_CONF )
        os.system( "%s %s" % ( os.environ.get('EDITOR', 'nano') , DRM4G_RESOURCES_CONF ) )
        self.check( )

    def list( self ) :
        """
        Check if the resource.conf file has been configured well and list the resources available.
        """
        self.check( )
        console_logger.info( "\033[1;4m%-20.20s%-20.20s\033[0m" % ('RESOURCE', 'STATE' ) )
        for resname, resdict in sorted( self.config.resources.items() ) :
            # To ignore VMs created by the DRM4G, since they should appear as hosts, not resources
            if '_' in resname:
                first_half, _ = resname.rsplit('_',1)
                if first_half in self.config.resources.keys() :
                    continue
            if resdict[ 'enable' ] == 'true' :
                state = 'enabled'
            else :
                state = 'disabled'
            console_logger.info( "%-20.20s%s" % ( resname, state ) )

    def features( self ) :
        """
        List the features of a given resource.
        """
        self.check( )
        for resname, resdict in sorted( self.config.resources.items() ) :
            console_logger.info( "Resource '%s' :" % ( resname ) )
            for key , val in sorted( resdict.items() ) :
                console_logger.info( " --> '%s' : '%s'" % ( key, val ) )

    def check( self ) :
        """
        Check if the resource.conf file has been configured well.
        """
        self.config.load()
        errors = self.config.check()
        if errors :
            raise Exception( "Please, review your configuration file:\n%s" % pprint.pformat(errors) )

class Proxy( object ):

    def __init__( self, resource, communicator ):
        self.resource     = resource
        self.communicator = communicator
        if 'myproxy_server' in self.resource :
            self.prefix = "X509_USER_PROXY=%s MYPROXY_SERVER=%s %s" % (
                                                                 join( REMOTE_VOS_DIR , self.resource[ 'myproxy_server' ] ),
                                                                 self.resource[ 'myproxy_server' ],
                                                                 "GT_PROXY_MODE=rfc " if self.resource[ "lrms" ] == "rocci" else ""
                                                                 )
        else :
            self.prefix = "X509_USER_PROXY=%s/${MYPROXY_SERVER} %s" % ( REMOTE_VOS_DIR,
                                                                        "GT_PROXY_MODE=rfc " if self.resource[ "lrms" ] == "rocci" else "" )

    def create( self , proxy_lifetime ):
        console_logger.info("--> Creating '%s' directory to store the proxy ... " % REMOTE_VOS_DIR )
        cmd = "mkdir -p %s" % REMOTE_VOS_DIR
        console_logger.debug( "Executing command ... " + cmd )
        out, err = self.communicator.execCommand( cmd )
        if not err :
            console_logger.info("--> Create a local proxy credential ... ")
            message      = 'Insert your Grid password: '
            grid_passwd  = getpass.getpass(message)
            cmd = self.prefix + "myproxy-init -S --cred_lifetime %s --proxy_lifetime %s --local_proxy -n -d" % (
                                                                                                         proxy_lifetime ,
                                                                                                         proxy_lifetime
                                                                                                         )
            console_logger.debug( "Executing command ... " + cmd )
            out , err = self.communicator.execCommand( cmd , input = grid_passwd )
            console_logger.info( out )
            if err :
                console_logger.info( err )
        else :
            raise Exception( err )

    def configure( self ) :
        certificate = self.resource.get( 'grid_cert' )
        if not certificate :
            console_logger.warning( " WARNING: It is assumed that the grid certificate has been already configured" )
        else :
            dir_certificate   = dirname( certificate )
            base_certificate  = basename( certificate )
            console_logger.info( "--> Converting '%s' key to pem format ... " % base_certificate )
            cmd = "openssl pkcs12 -nocerts -in %s -out %s" % ( certificate, join( dir_certificate, 'userkey.pem' ) )
            out , err = exec_cmd( cmd, stdin=sys.stdin, stdout=sys.stdout )
            if "invalid password" in err :
                raise Exception( err )
            console_logger.info( "--> Converting '%s' certificate to pem format ... " % base_certificate )
            cmd = "openssl pkcs12 -clcerts -nokeys -in %s -out %s" % ( certificate, join( dir_certificate, 'usercert.pem' ) )
            out , err = exec_cmd( cmd , stdin=sys.stdin, stdout=sys.stdout )
            if "invalid password" in err :
                raise Exception( err )
            console_logger.debug( "--> Creating '~/.globus' directory ... " )
            cmd = "mkdir -p ~/.globus"
            console_logger.debug( "Executing command ... " + cmd )
            out, err = self.communicator.execCommand( cmd )
            if err :
                raise Exception( err )
            for file in [ 'userkey.pem' , 'usercert.pem' ] :
                cmd = "rm -rf $HOME/.globus/%s" % file
                console_logger.debug( "Executing command ... " + cmd )
                out, err = self.communicator.execCommand( cmd )
                if err :
                    raise Exception( err )
                console_logger.info( "--> Copying '%s' to '%s' ..." % ( file , self.resource.get( 'frontend' ) ) )
                self.communicator.copy( 'file://%s'  % join( dir_certificate, file ) ,
                                        'ssh://_/%s' % join( '.globus' , file ) )
            console_logger.info( "--> Modifying userkey.pem permissions ... " )
            cmd = "chmod 400 $HOME/.globus/userkey.pem"
            console_logger.debug( "Executing command ... " + cmd )
            out, err = self.communicator.execCommand( cmd )
            if err :
                raise Exception( err )
            console_logger.info( "--> Modifying usercert.pem permissions ... " )
            cmd = "chmod 644 $HOME/.globus/usercert.pem"
            console_logger.debug( "Executing command ... " + cmd )
            out, err = self.communicator.execCommand( cmd )
            if err :
                raise Exception( err )

    def check( self ):
        console_logger.info( "--> Display information about the proxy certificate" )
        cmd = self.prefix + "grid-proxy-info"
        console_logger.debug( "Executing command ... " + cmd )
        out, err = self.communicator.execCommand( cmd )
        console_logger.info( out )
        if err :
            console_logger.info( err )

    def destroy( self ):
        console_logger.info( "--> Remove grid credentials" )
        cmd = self.prefix + "myproxy-destroy"
        console_logger.debug( "Executing command ... " + cmd )
        out , err = self.communicator.execCommand( cmd )
        console_logger.info( out )
        if err :
            console_logger.info( err )
        cmd = self.prefix + "grid-proxy-destroy"
        console_logger.debug( "Executing command ... " + cmd )
        out , err = self.communicator.execCommand( cmd )
        console_logger.info( out )
        if err :
            console_logger.info( err )




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import time
import threading
import logging
from os.path                 import join, dirname
from drm4g.communicators     import REMOTE_JOBS_DIR
from drm4g.utils.rsl2        import Rsl2Parser
from drm4g.utils.list        import List
from drm4g.core.configure    import Configuration
from drm4g.utils.dynamic     import ThreadPool
from drm4g.utils.message     import Send


class GwEmMad (object):
    """
    Execution manager MAD

    GridWay uses a Middleware Access Driver (MAD) module to submit,
    control and monitor the execution of jobs.

    The format to send a request to the Execution MAD, through its
    standard input, is:
    OPERATION JID HOST/JM RSL

    Where:

    -OPERATION: Can be one of the following:
        -INIT: Initializes the MAD (i.e. INIT - - -).
        -SUBMIT: Submits a job(i.e. SUBMIT JID HOST/JM RSL).
        -POLL: Polls a job to obtain its state (i.e. POLL JID - -).
    -CANCEL: Cancels a job (i.e. CANCEL JID - -).
    -FINALIZE:Finalizes the MAD (i.e. FINALIZE - - -).
    -JID: Is a job identifier, chosen by GridWay.
    -HOST: If the operation is SUBMIT, it specifies the resource contact
        to submit the job. Otherwise it is ignored.
    -JM: If the operation is SUBMIT, it specifies the job manager to submit
        the job. Otherwise it is ignored.
    -RSL: If the operation is SUBMIT, it specifies the resource specification
        to submit the job. Otherwise it is ignored.

    The format to receive a response from the MAD, through its standard output, is:

    OPERATION JID RESULT INFO

         Where:

    -OPERATION: Is the operation specified in the request that originated
        the response or CALLBACK, in the case of an asynchronous notification
        of a state change.
    -JID: It is the job identifier, as provided in the submission request.
    -RESULT: It is the result of the operation. Could be SUCCESS or FAILURE
    -INFO: If RESULT is FAILURE, it contains the cause of failure. Otherwise,
        if OPERATION is POLL or CALLBACK,it contains the state of the job.
    """
    logger  = logging.getLogger(__name__)
    message = Send()

    def __init__(self):
        self._callback_interval = 30 #seconds
        self._max_thread        = 10
        self._min_thread        = 3
        self._job_list          = List()
        self._configure         = None
        self._communicators     = dict()
        self._lock              = threading.Lock()

    def do_INIT(self, args):
        """
        Initializes the MAD (i.e. INIT - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'INIT - SUCCESS -'
        self.message.stdout( out )
        self.logger.debug( out )

    def do_SUBMIT(self, args):
        """
        Submits a job(i.e. SUBMIT JID HOST/JM RSL).
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, HOST_JM, RSL = args.split()
        try:
            HOST, JM = HOST_JM.rsplit( '/', 1 )
            # Init Job class
            job, communicator = self._update_resource( HOST )
            job.Communicator  = communicator
            # Parse rsl
            rsl                = Rsl2Parser(RSL).parser()
            if 'project' in job.resfeatures :
                rsl['project']      = job.resfeatures[ 'project' ]
            if 'parallel_env' in job.resfeatures :
                rsl['parallel_env'] = job.resfeatures[ 'parallel_env' ]
            if 'vo' in job.resfeatures and "::" in HOST :
                _ , host                    = HOST.split('::')
                job.resfeatures['host']     = host
                job.resfeatures['jm']       = JM
                job.resfeatures['env_file'] = join( dirname(RSL), "job.env" )
                job.resfeatures['queue']    = rsl[ 'queue' ]
            # Update remote directories
            ABS_REMOTE_JOBS_DIR = job.get_abs_directory( job.resfeatures.get( 'scratch', REMOTE_JOBS_DIR ) )
            for key in [ "stdout" , "stderr" , "executable" ] :
                rsl[key] = join( ABS_REMOTE_JOBS_DIR , rsl[key] )
            # Create and copy wrapper_drm4g file
            local_file    = join( dirname ( RSL ), "wrapper_drm4g.%s" % RSL.split( '.' )[ -1 ] )
            remote_file   = join( dirname ( rsl[ 'executable' ] ), 'wrapper_drm4g' )
            job.createWrapper( local_file, job.jobTemplate( rsl ) )
            job.copyWrapper( local_file, remote_file )
            # Execute wrapper_drm4g
            job.JobId = job.jobSubmit( remote_file )
            self._job_list.put( JID, job )
            out = 'SUBMIT %s SUCCESS %s:%s' % ( JID, HOST, job.JobId )
        except Exception as err:
            out = 'SUBMIT %s FAILURE %s' % ( JID, str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout(out)
        self.logger.debug(out)

    def do_FINALIZE(self, args):
        """
        Finalizes the MAD (i.e. FINALIZE - - -).
        @param args : arguments of operation
        @type args : string
        """
        out = 'FINALIZE - SUCCESS -'
        self.message.stdout( out )
        self.logger.debug( out )
        sys.exit( 0 )

    def do_POLL(self, args):
        """
        Polls a job to obtain its state (i.e. POLL JID - -).
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, HOST_JM, RSL = args.split()
        try:
            if self._job_list.has_key( JID ) :
                job    = self._job_list.get( JID )
                status = job.getStatus( )
                out = 'POLL %s SUCCESS %s' % ( JID, status )
            else:
                out = 'POLL %s FAILURE Job not submitted' % ( JID )
        except Exception as err:
            out = 'POLL %s FAILURE %s' % ( JID, str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_RECOVER(self, args):
        """
        Polls a job to obtain its state (i.e. RECOVER JID - -).
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, HOST_JM, RSL = args.split()
        try:
            HOST, remote_job_id = HOST_JM.split( ':', 1 )
            job , communicator  = self._update_resource( HOST )
            job.Communicator    = communicator
            job.JobId           = remote_job_id
            job.refreshJobStatus( )
            self._job_list.put( JID, job )
            out = 'RECOVER %s SUCCESS %s' % ( JID, job.getStatus( ) )
        except Exception as err:
            out = 'RECOVER %s FAILURE %s' % ( JID, str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_CALLBACK(self):
        """
        Show the state of the job
        """
        while True:
            time.sleep( self._callback_interval )
            self.logger.debug( "CALLBACK new iteration ..." )
            for JID , job in self._job_list.items( ):
                try:
                    self.logger.debug( "CALLBACK checking job '%s'" % JID  )
                    oldStatus = job.getStatus( )
                    job.refreshJobStatus( )
                    newStatus = job.getStatus( )
                    if oldStatus != newStatus or newStatus == 'DONE' or newStatus == 'FAILED':
                        if newStatus == 'DONE' or newStatus == 'FAILED':
                            self._job_list.delete(JID)
                            time.sleep ( 0.1 )
                        out = 'CALLBACK %s SUCCESS %s' % ( JID, newStatus )
                        self.message.stdout( out )
                        self.logger.debug( out )
                except Exception as err:
                    out = 'CALLBACK %s FAILURE %s' % ( JID, str( err ) )
                    self.logger.error( err , exc_info=1 )
                    self.message.stdout( out )

    def do_CANCEL(self, args):
        """
        Cancels a job (i.e. CANCEL JID - -).
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, HOST_JM, RSL = args.split()
        try:
            if self._job_list.has_key( JID ) :
                self._job_list.get(JID).jobCancel()
                out = 'CANCEL %s SUCCESS -' % (JID)
            else:
                out = 'CANCEL %s FAILURE Job not submitted' % (JID)
        except Exception as err:
            out = 'CANCEL %s FAILURE %s' % ( JID, str(err) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug( out )

    methods = {'INIT'    : do_INIT,
               'SUBMIT'  : do_SUBMIT,
               'POLL'    : do_POLL,
               'RECOVER' : do_RECOVER,
               'CANCEL'  : do_CANCEL,
               'FINALIZE': do_FINALIZE}

    def processLine(self):
        """
        Choose the OPERATION through the command line
        """
        try:
            worker = threading.Thread( target = self.do_CALLBACK, )
            worker.setDaemon( True )
            worker.start()
            self._configure = Configuration()
            pool = ThreadPool( self._min_thread, self._max_thread )
            while True:
                input = sys.stdin.readline().split()
                self.logger.debug( ' '.join(input) )
                OPERATION = input[0].upper()
                if len(input) == 4 and OPERATION in self.methods:
                    if OPERATION in ( 'FINALIZE', 'INIT', 'SUBMIT', 'RECOVER' ):
                        self.methods[ OPERATION ]( self, ' '.join(input) )
                    else:
                        pool.add_task( self.methods[ OPERATION ], self, ' '.join(input) )
                else:
                    out = 'WRONG COMMAND'
                    self.message.stdout( out )
                    self.logger.debug( out )
        except Exception as err:
            self.logger.warning( str ( err ) , exc_info=1 )

    def _update_resource(self, host):
        with self._lock :
            if not self._configure.check_update() or not self._configure.resources :
                self._configure.load()
                errors = self._configure.check()
                if errors :
                    self.logger.error ( ' '.join( errors ) )
                    raise Exception ( ' '.join( errors ) )
            for resname, resdict in list( self._configure.resources.items() ) :
                if  'cloud_provider' in self._configure.resources[ resname ].keys():
                    continue
                if '::' in host :
                    _resname , _ = host.split( '::' )
                    if resname != _resname :
                        continue
                elif resname != host :
                    continue
                if resname not in self._communicators :
                    self._communicators[ resname ] = self._configure.make_communicators()[resname]
                job          = self._configure.make_resources()[ resname ]['Job']
                communicator = self._communicators[ resname ]
                return job, communicator





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

__all__ = ["communicators", "core", "handlers", "utils"]




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#


scheme   = None
netloc   = None
path     = None
params   = { }
query    = None
fragment = None

def urlparse (url):
    """
    Parse a URL and return a ParseResult object
    """
    global scheme, netloc, path, params, query, fragment
    if '://' in url:
        _scheme, url = url.split('://', 1)
        scheme = _scheme.lower()
    if '#' in url:
        url, fragment = url.split('#', 1)
    if '?' in url:
        url, query  = url.split('?', 1)
        params = dict(( elem.split('=')[0], elem.split('=')[1]) for elem in query.split(';') if '=' in elem)
    if '/' in url:
        if scheme == 'file':
            path = url
        else:
            url, path  = url.split('/', 1)
    netloc = url
    return ParseResult(scheme, netloc, path, params, query, fragment)

class ParseResult(object):
    """
    Methods for the parsed result objects
    """
    def __init__(self, scheme, netloc, path, params, query, fragment):

        self._scheme   = scheme
        self._netloc   = netloc
        self._path     = path
        self._params   = params
        self._query    = query
        self._fragment = fragment

    def getscheme(self):
        return self._scheme

    def getnetloc(self):
        return self._netloc

    def getpath(self):
        return self._path

    def getparams(self):
        return self._params

    def getquery(self):
        return self._query

    def getfragment(self):
        return self._fragment

    def getusername(self):
        netloc = self._netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                userinfo = userinfo.split(":", 1)[0]
            return userinfo
        return None

    def getpassword(self):
        netloc = self._netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                return userinfo.split(":", 1)[1]
        return None

    def gethost(self):
        netloc = self._netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            netloc = netloc.split(":", 1)[0]
        return netloc or None

    def getport(self):
        netloc = self._netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            port = netloc.split(":", 1)[1]
            return int(port, 10)
        return None

    scheme   = property(getscheme)
    netloc   = property(getnetloc)
    path     = property(getpath)
    params   = property(getparams)
    query    = property(getquery)
    fragment = property(getfragment)
    username = property(getusername)
    password = property(getpassword)
    host     = property(gethost)
    port     = property(getport)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
from os.path     import dirname, abspath, join, expanduser, exists

try:
    sys.path.insert( 0, join( dirname( dirname ( abspath( __file__ ) ) ), 'utils' ) )
    from paramiko.transport     import Transport
    from paramiko.agent         import Agent
    from paramiko.dsskey        import DSSKey
    from paramiko.rsakey        import RSAKey
    from scp                    import SCPClient
except Exception as e:
    exit( 'Caught exception: %s' % str(e) )

import re
import socket
import drm4g.commands
import drm4g.communicators
from drm4g.communicators    import ComException, SSH_CONNECT_TIMEOUT, SFTP_CONNECTIONS
from drm4g.utils.url        import urlparse

import logging
logger  = logging.getLogger(__name__)

class Communicator(drm4g.communicators.Communicator):
    """
    Create a SSH session to remote resources.
    """
    _lock       = __import__('threading').Lock()
    _sem        = __import__('threading').Semaphore(SFTP_CONNECTIONS)
    _trans      = None

    def connect(self):
        try:
            with self._lock :
                if not self._trans or not self._trans.is_authenticated( ) :
                    logger.debug("Opening ssh connection ... ")
                    keys = None
                    logger.debug("Trying ssh-agent ... " )
                    drm4g_agent = drm4g.commands.Agent()
                    drm4g_agent.start()
                    drm4g_agent.update_agent_env()
                    # paramiko agent
                    agent = Agent()
                    keys  = agent.get_keys()
                    if not keys :
                        logger.debug( "Error trying to connect to '%s'" % self.frontend )
                        logger.debug( "Impossible to load '%s' key from the ssh-agent"  % self.private_key )
                        try:
                            status_ssh_agent = agent._conn
                        except Exception as err :
                            logger.warning( "Probably you are using paramiko version <= 1.7.7.2 : %s " % err )
                            status_ssh_agent = agent._conn
                        if not status_ssh_agent:
                            logger.warning( "'ssh-agent' is not running" )
                        else:
                            if agent.get_keys():
                                logger.warning( "ssh-agent is running but none of the keys have been accepted"
                                "by remote frontend %s." % self.frontend )
                            else:
                                logger.debug( "'ssh-agent' is running but without any keys" )
                    if self.private_key :
                        logger.debug("Trying '%s' key ... " % self.private_key )
                        private_key_path = expanduser( self.private_key )
                        if ( not exists( private_key_path ) ) and ( not 'PRIVATE KEY' in  self.private_key ):
                            output = "'%s'key does not exist" % private_key_path
                            raise ComException( output )
                        for pkey_class in (RSAKey, DSSKey):
                            try :
                                if 'PRIVATE KEY' in self.private_key :
                                    import StringIO
                                    key  = pkey_class.from_private_key( StringIO.StringIO ( self.private_key.strip( "'" ) ) )
                                else :
                                    key  = pkey_class.from_private_key_file( private_key_path )
                                keys = keys + (key,)
                            except Exception :
                                pass
                    if not keys :
                        output = "Impossible to load any keys"
                        logger.error( output )
                        raise ComException( output )

                    for key in keys:
                        try:
                            sock = socket.socket()
                            try:
                                sock.settimeout( SSH_CONNECT_TIMEOUT )
                            except :
                                output = "Timeout trying to connect to '%s'" % self.frontend
                                raise ComException( output )
                            logger.debug( "Connecting to '%s' as user '%s' port  '%s' ..."
                                               % ( self.frontend , self.username, self.port ) )
                            if ':' in self.frontend :
                                self.frontend , self.port = self.frontend.split( ':' )
                            sock.connect( ( self.frontend , self.port ) )
                            self._trans      = Transport( sock )
                            self._trans.connect( username = self.username , pkey = key )
                            if self._trans.is_authenticated( ) :
                                break
                        except socket.gaierror:
                            output = "Could not resolve hostname '%s' " % self.frontend
                            raise ComException( output )
                        except Exception as  err :
                            logger.warning( "Error connecting '%s': %s" % ( self.frontend , str ( err ) ) )
                if not self._trans :
                    output = "Authentication failed for '%s'. Try to execute `ssh -vvv -p %d %s@%s` and see the response." % (
                              self.frontend , self.port, self.username, self.frontend )
                    raise ComException( output  )
        except ComException:
            raise
        except Exception as err:
            if "No handlers could be found for logger" in str(err):
                raise Exception("The connect function is the one causing problems : %s" % str(err))
            else:
                raise

    def execCommand(self , command , input=""):
        self.connect()
        with self._lock :
            channel = self._trans.open_session()
        channel.settimeout( SSH_CONNECT_TIMEOUT )
        channel.exec_command( command )
        if input :
            for line in input.split( ):
                channel.makefile( 'w').write( '%s\n' % line )
                channel.makefile( 'w' ).flush( )
        stdout = ''.join( channel.makefile( 'r').readlines( ) )
        stderr = ''.join( channel.makefile_stderr( 'r').readlines( ) )
        if channel :
            channel.close( )
        return stdout , stderr

    def mkDirectory(self, url):
        to_dir         = self._set_dir(urlparse(url).path)
        stdout, stderr = self.execCommand( "mkdir -p %s" % to_dir )
        if stderr :
            raise ComException( "Could not create %s directory on '%s': %s" % ( to_dir , self.frontend , stderr ) )

    def rmDirectory(self, url):
        to_dir         = self._set_dir(urlparse(url).path)
        stdout, stderr = self.execCommand( "rm -rf %s" % to_dir )
        if stderr:
            raise ComException( "Could not remove %s directory on '%s': %s" % ( to_dir , self.frontend , stderr ) )

    def copy( self , source_url , destination_url , execution_mode = '' ) :
        with self._sem :
            self.connect( )
            scp = SCPClient( self._trans )
            if 'file://' in source_url :
                from_dir = urlparse( source_url ).path
                to_dir   = self._set_dir( urlparse( destination_url ).path )
                logger.debug( "Putting '%s' -> '%s'" %  (from_dir, to_dir  ))
                scp.put( from_dir , to_dir )
                if execution_mode == 'X':
                    stdout, stderr = self.execCommand( "chmod +x %s" % to_dir )
            else:
                from_dir = self._set_dir( urlparse( source_url ).path )
                to_dir   = urlparse(destination_url).path
                logger.debug( "Getting '%s' -> '%s'" %  (from_dir, to_dir  ))
                scp.get( from_dir, to_dir )


    def close( self ) :
        try :
            if self._trans :
                self._trans.close( )
        except Exception as err:
            logger.warning( "Could not close the SSH connection to '%s': %s" % ( self.frontend , str( err ) ) )

    def __del__( self ) :
        """
        Attempt to clean up if not explicitly closed.
        """
        self.close( )


    #internal
    def _set_dir(self, path):
        work_directory =  re.compile( r'^~' ).sub( self.work_directory , path )
        if work_directory[0] == r'~' :
            return ".%s" %  work_directory[ 1: ]
        else :
            return  work_directory



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage identities for resources. That involves managing private/public keys
and grid credentials, depending on the resource configuration.

Usage:
    drm4g id <resource_name> init   [ options ] [ --lifetime=<hours> ]
    drm4g id <resource_name> info   [ options ]
    drm4g id <resource_name> delete [ options ]

 Options:
    -l --lifetime=<hours>   Duration of the identity's lifetime [default: 168].
    -d --debug              Debug mode.

Commands:

    init                    Create an identity for a certain period of time, by
                            default 168 hours (1 week). Use the option --lifetime
                            to modify this value. It adds the configured private
                            key to a ssh-agent and creates a grid proxy using
                            myproxy server.
                            It appends the public key to the remote user's
                            ~/.ssh/authorized_keys file (creating the file, and
                            directory, if necessary). It tries to load the public
                            key obtained by appending *.pub to the name of the
                            configured private key file. Alternative the public
                            key can be given by public_key variable.
                            It also configures the user's grid certificate
                            under ~/.globus directory (creating directory,
                            if necessary) if grid_cert variable is defined.

    info                    Get some information about the identity's status.

    delete                  Remove the identity from the ssh-agent and the
                            myproxy server.
"""

from drm4g.core.configure import Configuration
from drm4g.commands       import Daemon, Agent, Proxy
from drm4g                import console_logger

def run( arg ) :
    try :
        daemon = Daemon()
        if not daemon.is_alive() :
            raise Exception( 'DRM4G is stopped.' )
        config = Configuration()
        config.load( )
        if config.check( ) :
            raise Exception( "Review the configuration of '%s'." % ( arg['<resource_name>'] ) )
        if arg['<resource_name>'] not in config.resources :
            raise Exception( "'%s' is not a configured resource." % ( arg['<resource_name>'] ) )
        lrms         = config.resources.get( arg['<resource_name>'] )[ 'lrms' ]
        communicator = config.resources.get( arg['<resource_name>'] )[ 'communicator' ]
        if lrms != 'cream' and lrms != 'rocci' and communicator == 'local' :
            raise Exception( "'%s' does not have an identity to configure." % ( arg['<resource_name>'] ) )
        if lrms == 'cream' or lrms == 'rocci' :
            comm = config.make_communicators()[ arg['<resource_name>'] ]
            if communicator == 'op_ssh' :
                #paramiko will always be used to renew the grid certificate
                config.resources.get( arg['<resource_name>'] )[ 'communicator' ] = 'pk_ssh'
                comm = config.make_communicators()[ arg['<resource_name>'] ]
            proxy = Proxy( config.resources[ arg['<resource_name>'] ] ,
                           comm
                           )
            config.resources.get( arg['<resource_name>'] )[ 'communicator' ] = communicator
            config.make_communicators()
        if communicator != 'local' :
            agent = Agent( config.resources[ arg['<resource_name>'] ] )
        if arg[ 'init' ] :
            if communicator != 'local' :
                agent.start( )
                agent.add_key( arg[ '--lifetime' ] )
                agent.copy_key( )
            if lrms == 'cream' or lrms == 'rocci' :
                proxy.configure( )
                proxy.create( arg[ '--lifetime' ] )
        elif arg[ 'delete' ] :
            if lrms == 'cream' or lrms == 'rocci' :
                proxy.destroy( )
            if communicator != 'local' :
                agent.delete_key( )
        else :
            if communicator != 'local' :
                agent.list_key( )
            if lrms == 'cream' or lrms == 'rocci' :
                proxy.check( )
    except Exception as err :
        console_logger.error( str( err ) )
